/***************************************************************************
                          xsltransformation.cpp -  description
                             -------------------
    begin                : Sat Oct 8 2005
    copyright            : (C) 2005 by Diederik van der Boor
    email                : vdboor --at-- codingdomain.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/*
 * The kopetexsl.cpp file from Kopete is used as working
 * demo/example to create this code, since it's more
 * to-the-point then the online libxml documentation.
 */

#include "xsltransformation.h"
#include "../kmessdebug.h"

#include <qfile.h>
#include <qcstring.h>
#include <qregexp.h>

#include <kurl.h>
#include <kglobal.h>
#include <kstddirs.h>

#include <libxml/globals.h>
#include <libxml/parser.h>

#include <libxslt/xsltconfig.h>
#include <libxslt/xsltInternals.h>
#include <libxslt/transform.h>  // After xsltconfig and xsltInternals or it breaks compiling on some libxslt versions.

// stdlib.h is required to build on Solaris
#include <stdlib.h>
#include <string.h>



// The constructor
XslTransformation::XslTransformation()
  : styleSheet_(0)
  , xslDoc_(0)
  , xslParams_(0)
{
  // Init libxml
  xmlLoadExtDtdDefaultValue = 0;
  xmlSubstituteEntitiesDefault( 1 );
}



// The destructor
XslTransformation::~XslTransformation()
{
  if(styleSheet_ != 0)
  {
    xsltFreeStylesheet(styleSheet_);
    // also frees xslDoc_
  }

  delete[] xslParams_;

  xsltCleanupGlobals();
  xmlCleanupParser();
}



// Convert an XML string
QString XslTransformation::convertXmlString(const QString &xml) const
{
  // Asserts
  if(KMESS_NULL(styleSheet_)) return QString::null;

  // Convert data to C-string
  QCString xmlCString = xml.utf8();

  // Create XmlDoc
  xmlDocPtr sourceXmlDoc = xmlParseMemory( xmlCString.data(), xmlCString.length() );
  if(sourceXmlDoc == 0)
  {
    kdWarning() << "XslTransformation: could not parse the XML data." << endl;
    return QString::null;
  }

  // Apply XSL
  xmlDocPtr resultXmlDoc = xsltApplyStylesheet( styleSheet_, sourceXmlDoc, xslParams_ );
  if(resultXmlDoc == 0)
  {
    kdWarning() << "XslTransformation: could not convert the XML data." << endl;
    // Free intermediate results. You gotta love C-API's.. ;-)
    xmlFree(sourceXmlDoc);
    return QString::null;
  }

  // Save
  xmlChar *xmlData     = 0;  // Filled by xmlDocDumpMemory()
  int      xmlDataSize = 0;
  xmlDocDumpMemory( resultXmlDoc, &xmlData, &xmlDataSize );
  QString resultXmlString = QString::fromUtf8( reinterpret_cast<char*>(xmlData), xmlDataSize );
  xmlFree( xmlData );        // Free data from xmlDocDumpMemory() again

  // Free XmlDoc 
  xmlFree(resultXmlDoc);
  xmlFree(sourceXmlDoc);

  // Strip xml header and last \n from result string
  resultXmlString = resultXmlString.replace( QRegExp("^<\\?xml [^?]+\\?>\n?"), QString::null )
                                   .replace( QRegExp("\r?\n?$"),               QString::null );

  // xsltApplyStylesheet returns an empty result when the parsing of user parameters failed.
  if( resultXmlString.isEmpty() )
  {
    return QString::null;
  }
  else
  {
    return resultXmlString;
  }
}



// Check whether a stylesheet is loaded successfully
bool XslTransformation::hasStylesheet() const
{
  return (styleSheet_ != 0);
}



// Set the XSL parameters
void XslTransformation::setParameters( const QMap<QString,QString> xslParameters )
{
  // Delete previous code.
  delete[] xslParams_;

  // Params
  if( xslParameters.isEmpty() )
  {
    xslParams_ = 0;
  }
  else
  {
    int argCount = xslParameters.count() * 2 + 1;
    xslParams_ = new const char*[ argCount ];

    int i = 0;
    for( QMapConstIterator<QString,QString> it = xslParameters.begin(); it != xslParameters.end(); ++it )
    {
      // Duplicate the strings as they will be deleted outside this scope.
      xslParams_[ i++ ] = strdup( it.key().utf8().data() );
      xslParams_[ i++ ] = strdup( QString( "'" + it.data() + "'" ).utf8().data() );
    }
    xslParams_[i++] = 0;

#ifdef KMESSTEST
    ASSERT( i == argCount );
#endif
  }
}



// Set the XSL stylesheet.
void XslTransformation::setStylesheet(const QString &xslFileName)
{
  // No expensive operations when file is identical.
  if(xslFileName == xslFileName_) return;
  xslFileName_ = xslFileName;

  // Clean up previous stylesheet
  if(styleSheet_ != 0)
  {
    xsltFreeStylesheet(styleSheet_);
    // also frees xslDoc_
    styleSheet_ = 0;
    xslDoc_     = 0;
  }

  // Get file path
  KURL fileUrl(xslFileName);
  QCString basePath = fileUrl.directory(false).utf8();

  // Open XSL file
  QFile file(xslFileName);
  if( ! file.open(IO_ReadOnly) )
  {
    kdWarning() << "XslTransformation: could not load '" << xslFileName << "'!" << endl;
    return;
  }

  // Read all data
  QByteArray fileData = file.readAll();
  file.close();

  // Parse file data as XML
  xslDoc_ = xmlParseMemory( fileData.data(), fileData.size() );
  if(xslDoc_ == 0)
  {
    kdWarning() << "XslTransformation: could not parse XML from '" << xslFileName << "'!" << endl;
    return;
  }

  // Set base path, required for xsl imports/includes
  xmlNodeSetBase( reinterpret_cast<xmlNode*>(xslDoc_),
                  reinterpret_cast<const xmlChar*>( basePath.data() ) );

  // Parse XML data as XSL
  styleSheet_ = xsltParseStylesheetDoc( xslDoc_ );
  if( styleSheet_ == 0 || styleSheet_->errors != 0 )
  {
    // Also happens when an <xsl:import> fails
    kdWarning() << "XslTransformation: could not parse XSL from '" << xslFileName << "'!" << endl;
    xmlFreeDoc( xslDoc_ );
    xslDoc_ = 0;
  }
}



