/***************************************************************************
                          applicationlist.h -  description
                             -------------------
    begin                : Mon 01 16 2005
    copyright            : (C) 2005 by Diederik van der Boor
    email                : "vdboor" --at-- "codingdomain.com"
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef APPLICATIONLIST_H
#define APPLICATIONLIST_H

#include <qobject.h>
#include <qptrlist.h>

class Application;
class MimeApplication;
class P2PApplication;
class P2PMessage;
class MimeMessage;
class ChatMessage;
class MsnObject;
class DirectConnectionPool;
class MsnDirectConnection;
class MsnSwitchboardConnection;



/**
 * @brief The collection of applications and direct connections associated with a contact.
 *
 * Peer-to-peer applications (P2PApplication instances) are created at the switchboard server,
 * but they are not limited to one chat session. Clients are allowed to use any active switchboard
 * or direct connection to deliver MSNP2P packets to the peer contact. A direct connection is even
 * re-used by other invitations with the same contact, for both new and existing invitations!
 * The initialized p2p applications remain active through the direct connection after the switchboard is closed.
 *
 * These design choices of the MSN Messenger protocol makes the client implementations heavier,
 * but it makes more efficient use of network resources.
 * After all, MSNC* appears to be all about moving protocol logic to the clients.
 * This class maintains the list of applications and connections for one contact.
 *
 * @author Diederik van der Boor
 * @ingroup Applications
 */
class ApplicationList : public QObject
{
  Q_OBJECT

  public:
    // The constructor
                           ApplicationList(const QString &contactHandle);
    // The destructor
    virtual               ~ApplicationList();

    // Add a new application to the list.
    void                   addApplication(MimeApplication *application);
    // Add a new application to the list.
    void                   addApplication(P2PApplication *application);
    // Attempt to establish a direct connection at the given ipaddress/port.
    bool                   addConnection(const QString &ipAddress, const int port);
    // Attempt to establish a direct connection by listening at an random port.
    int                    addServerConnection();
    // Abort all applications, the contact is leaving a chat.
    bool                   contactLeavingChat(const MsnSwitchboardConnection *connection, bool userInitiated = false);
    // Abort all applications, the contact left a chat.
    bool                   contactLeftChat(bool userInitiated = false);
    // Find the application with uses the given cookie.
    Application *          getApplicationByCookie(const QString &cookie) const;
    // Return the handle of the contact this class manages.
    const QString &        getContactHandle() const;
    // Deliver a message to the correct application instance
    void                   gotMessage(const MimeMessage &message);
    // Deliver a message to the correct application instance
    void                   gotMessage(const P2PMessage &message);
    // Return the direct connection, if available.
    MsnDirectConnection *  getDirectConnection() const;
    // Return whether there is an active direct connection.
    bool                   hasDirectConnection() const;
    // Return whether an MsnObject transfer for the given msn object is already running.
    bool                   hasMsnObjectTransfer( const MsnObject &msnObject ) const;
    // Return whether the transfer invitation has been sent.
    bool                   hasPendingConnectionInvitation() const;
    // Return whether there are pending connection attempts.
    bool                   hasPendingConnections() const;
    // Pause all applications, avoid flooding the switchboard message queue.
    void                   pauseApplications();
    // Resume all applications, allowing messages to be sent again.
    void                   resumeApplications(bool isPrivateChat = true);
    // Send a message for a P2P application over one of the available connections/bridges.
    void                   sendMessage(const P2PApplication *source, const QByteArray &header, const QByteArray &messageData = 0, uint footerCode = 0);
    // Set whether the transfer invitation has been sent.
    void                   setPendingConnectionInvitation(bool state);

  public slots:
    // Send a message for a Mime application.
    void                   sendMessage(const MimeApplication *source, const MimeMessage &messageData);

  private:  // private methods
    // Abort all applications, the connection is closing or closed
    bool                   abortApplications(const MsnSwitchboardConnection *abortingConnection = 0, bool userInitiated = false);
    // Connect the application object signals.
    void                   connectApplication(Application *app);
    // Create the application instance for the given invitation message.
    MimeApplication *      createApplication(const MimeMessage &message);
    // Create the application instance for the given invitation message.
    P2PApplication *       createApplication(const P2PMessage &message);
    // Create a P2PApplication instance for the given Application-GUID.
    MimeApplication *      createApplicationByGuid(const QString &applicationGuid);
    // Create a P2PApplication instance for the given EUF-GUID.
    P2PApplication *       createApplicationByEufGuid(const QString &eufGuid);
    // Find the application which should receive the ACK message.
    P2PApplication *       getApplicationByAckData(const P2PMessage &ackMessage) const;
    // Find the application which maintains the session for the given CallID.
    P2PApplication *       getApplicationByCallId(const QString &callID) const;
    // Find the application which received the previous message fragment from the contact.
    P2PApplication *       getApplicationByLastMessage(unsigned long messageID) const;
    // Find the application which authorizes the transfer for the given nonce.
    P2PApplication *       getApplicationByNonce(const QString &nonce) const;
    // Find the application identified with the given session ID.
    P2PApplication *       getApplicationBySessionId(unsigned long sessionID) const;
    // Create the direct connection pool
    void                   initializeDirectConnectionPool();
    // Reject an invitation for a mime-application.
    void                   sendMimeRejectMessage( const QString &invitationCookie, bool notInstalled );

  private slots:
    // The direct connection is authorized.
    void                   slotActiveConnectionAuthorized();
    // The direct connection was closed.
    void                   slotActiveConnectionClosed();
    // All direct connection attempts failed.
    void                   slotAllConnectionsFailed();
    // The direct connection is established.
    void                   slotConnectionEstablished();
    // A message was received from the direct connection.
    void                   slotGotDirectConnectionMessage(const QByteArray &messageData);
    // An application was finished and requested removal.
    void                   slotTerminateApplication(Application *application);


  private:  // private attributes
    // A list of all applications that are being aborted.
    QPtrList<Application>  abortingApplications_;
    // The contact this class was initiated for.
    QString                contactHandle_;
    // The direct connection established with the contact
    MsnDirectConnection   *directConnection_;
    // The list of direct connection attempts
    DirectConnectionPool  *directConnectionPool_;
    // A list of pointers to all open mime transfers.
    QPtrList<MimeApplication>  mimeApplications_;
    // A list of pointers to all open p2p transfers.
    QPtrList<P2PApplication>  p2pApplications_;
    // The direct connection invitation is sent.
    bool                   pendingConnectionInvitation_;

  signals:
    // Indicate the applications were aborted
    void                   applicationsAborted(const QString &contact);
    // The direct connection was authorized
    void                   connectionAuthorized();
    // The direct connection was closed
    void                   connectionClosed();
    // The direct connection was established.
    void                   connectionEstablished();
    // The direct connection attempts failed
    void                   connectionFailed();
    // A new application was initialized, connect it's signals
    void                   newApplication(Application *application);
    // Deliver a new message to the switchboard
    void                   putMsg(const MimeMessage &message, const QString &contactHandle, bool privateChatRequired);
};

#endif
