/***************************************************************************
                          p2papplication.cpp -  description
                             -------------------
    begin                : Mon Nov 22 2004
    copyright            : (C) 2004 by Diederik van der Boor
    email                : vdboor --at-- codingdomain.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "p2papplication.h"

#include <stdlib.h>
#include <string.h>  // for memcpy

#include <qregexp.h>
#include <qptrlist.h>
#include <qdatetime.h>
#include <qiodevice.h>
#include <qtimer.h>
#include <qbuffer.h>
#include <klocale.h>
#include <kdebug.h>

#include "applicationlist.h"
#include "../mimemessage.h"
#include "../p2pmessage.h"
#include "../extra/msndirectconnection.h"
#include "../../msnobject.h"
#include "../../kmessdebug.h"
#include "../../currentaccount.h"

#ifdef KMESSDEBUG_P2PAPPLICATION
#define KMESSDEBUG_P2PAPPLICATION_GENERAL
#endif


/**
 * @brief Constructor, initializes the P2PApplication instance fields.
 * @param  applicationList  The application list object for the contact.
 *
 * The ApplicationList object contains all shared resources
 * for P2P applications of one contact. This object manages the
 * available connections to send the P2P messages.
 */
P2PApplication::P2PApplication(ApplicationList *applicationList)
: Application( applicationList->getContactHandle() ),
  aborting_(false),
  applicationList_(applicationList),
  buffer_(0),
  dataSource_(0),
  dataType_(P2P_TYPE_NEGOTIATION),
  gotSlpMessage_(false),
  invitationCSeq_(0),
  invitationSessionID_(0),
  sessionID_(0),
  messageID_(0),
  messageOffset_(0),
  messageTotalSize_(0),
  shouldSendAck_(false),
  userShouldAcknowledge_(false),
  waitingState_(P2P_WAIT_DEFAULT)
{
  // Reset other internal fields
  lastIncomingMessage_.dataSize     = 0;
  lastIncomingMessage_.messageID    = 0;
  lastIncomingMessage_.messageType  = P2P_MSG_UNKNOWN;
  lastIncomingMessage_.sentTime     = 0;
  lastIncomingMessage_.sessionID    = 0;
  lastIncomingMessage_.totalSize    = 0;
  lastIncomingMessage_.uniqueID     = 0;

  // Enable auto-deletion on ack queue properties.
  incomingMessages_.setAutoDelete(true);
  outgoingMessages_.setAutoDelete(true);

  // Initiaize the timer
  waitingTimer_ = new QTimer(this);
  connect( waitingTimer_, SIGNAL(timeout()), this, SLOT(slotCleanup()) );
}



/**
 * Class destructor.
 */
P2PApplication::~P2PApplication()
{
  // Stop the timer
  waitingTimer_->stop();

  // Delete pointers
  delete waitingTimer_;
  delete buffer_;
}



/**
 * Step one of a contact-started chat: the contact invites the user
 * By default the invitation will be declined unless you override this method.
 */
void P2PApplication::contactStarted1_ContactInvitesUser(const MimeMessage &message)
{
#ifdef KMESSDEBUG_APPLICATION
  kdDebug() << "P2PApplication - contactStarted1_ContactInvitesUser (rejects invitation)" << endl;
#endif

  // Cancel the invitation
  sendCancelMessage( CANCEL_NOT_INSTALLED );

  // We can assume the switchboard already displayed a better error message
#ifdef KMESTEST
  kdWarning() << "The Application subclass didn't implement contactStarted1_ContactInvitesUser!" << endl;
#endif

  // Send some debug info to the console
  MimeMessage slpContent(message.getBody());
  kdWarning() << "The contact initiated a MSN6 feature KMess can't handle yet "
              << "(euf-guid=" << slpContent.getValue("EUF-GUID") << " appid=" << slpContent.getValue("AppID") << ")." << endl;

  // Display a warning to the user
  showSystemMessage( i18n("The contact initiated a MSN6 feature KMess can't handle yet.") );


  // Unlike Application::contactStarted1_ContactInvitesUser()
  // we don't quit here!! We should wait for the BYE message from the other contact!
}



/**
 * Step four of a contact-started chat: the contact confirms the data-preparation message.
 */
void P2PApplication::contactStarted4_ContactConfirmsPreparation()
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication - contactStarted4_ContactConfirmsPreparation" << endl;
#endif
}



/**
 * Private utility function to generate an Session and message/sequence ID.
 * The message/sequence ID starts random, between [4, 4294967295].
 */
unsigned long P2PApplication::generateID()
{
  return (rand() & 0x00FFFFFC);
}



/**
 * Private utility function to generate a Call ID or Branch ID.
 * This ID has the format of a GUID, but it can in fact by any random string.
 */
QString P2PApplication::generateGUID()
{
  // This code is based on Kopete, but much shorter.
  QString guid = "{"
      + QString::number((rand() & 0xAAFF) + 0x1111, 16)
      + QString::number((rand() & 0xAAFF) + 0x1111, 16)
      + "-"
      + QString::number((rand() & 0xAAFF) + 0x1111, 16)
      + "-"
      + QString::number((rand() & 0xAAFF) + 0x1111, 16)
      + "-"
      + QString::number((rand() & 0xAAFF) + 0x1111, 16)
      + "-"
      + QString::number((rand() & 0xAAFF) + 0x1111, 16)
      + QString::number((rand() & 0xAAFF) + 0x1111, 16)
      + QString::number((rand() & 0xAAFF) + 0x1111, 16)
      + "}";
  return guid.upper();
}



/**
 * Return the branch ID, this identifies the SLP INVITE-message.
 */
QString P2PApplication::getBranch() const
{
  return branch_;
}



/**
 * Return the call ID, this identifies the SLP session.
 */
QString P2PApplication::getCallID() const
{
  return callID_;
}



/**
 * Return the content type read from the invitation message
 * This method is useful for some contactStarted1_ContactInvitesUser() implementations.
 */
const QString& P2PApplication::getInvitationContentType() const
{
  return invitationContentType_;
}



/**
 * Return the Session ID this class read from the INVITE message.
 * This method is useful for some contactStarted2_UserAccepts() implementations.
 */
unsigned long P2PApplication::getInvitationSessionID() const
{
  return invitationSessionID_;
}



/**
 * Return the message ID used in the previous message received from the contact.
 * This method is used to match partial/fragmented messages.
 */
unsigned long P2PApplication::getLastContactMessageID() const
{
  return lastIncomingMessage_.messageID;
}



/**
 * Returns the nonce that's being expected.
 */
const QString & P2PApplication::getNonce() const
{
  return nonce_;
}



/**
 * Return the session ID, this identifies the P2P session.
 */
unsigned long P2PApplication::getSessionID() const
{
  return sessionID_;
}



/**
 * Find the unacked message record associated with the received P2P message.
 * @param  message  The ACK message to test for.
 */
P2PApplication::UnAckedMessage * P2PApplication::getUnAckedMessage(const P2PMessage &ackMessage) const
{
#ifdef KMESSTEST
  ASSERT( ackMessage.getFlags() != 0 && ackMessage.getDataSize() == 0 );
#endif

  unsigned long ackUniqueID  = ackMessage.getAckUniqueID();
  unsigned long ackMessageID = ackMessage.getAckMessageID();

  QPtrListIterator<UnAckedMessage> it(outgoingMessages_);
  while( it.current() != 0 )
  {
    UnAckedMessage *unAcked = it.current();
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
    kdDebug() << "P2PApplication: Testing if P2P ACK matches"
              << "    (UniqueID "  << unAcked->uniqueID  << " == " << ackUniqueID  << " ?)"
              << " or (UniqueID "  << unAcked->uniqueID  << " == " << ackMessageID << " ?)"
              << " or (MessageID " << unAcked->messageID << " == " << ackMessageID << " ?)" << endl;
#endif

    if(  unAcked->uniqueID == ackUniqueID   // This works with MSN 6.2
    ||  (unAcked->uniqueID == ackMessageID && ackUniqueID == 0) )  // For type-8 error messages
    {
      // Found it, save reference
      return unAcked;
    }

    // Also attempt to match on the messageID alone if the UniqueID is not set.
    if( ackUniqueID == 0 && unAcked->messageID == ackMessageID )
    {
      // HACK: added for GAIM 1.5 (does not return UniqueID)
      // Calling ApplicationList displays a warning, so it does not appear twice.
      return unAcked;
    }

    ++it;
  }

  return 0;
}



/**
 * Parse the ACK message. In certain situations we need to activate some methods.
 * @param  message      The ACK message.
 * @param  messageType  The original message type the ACK is for.
 */
void P2PApplication::gotAck(const P2PMessage & message, const P2PMessageType messageType)
{
  switch( messageType )
  {
    // Are we waiting for the contact to accept our invitation (e.g. send "200 OK")?
    case P2P_MSG_SESSION_INVITATION:

      // Don't change the waiting state yet
      waitingTimer_->start(30000, true); // 30 sec
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication: Got an ACK message (INVITE ACK, still waiting for 200 OK)." << endl;
#endif
#ifdef KMESSTEST
      ASSERT( waitingState_ == P2P_WAIT_CONTACT_ACCEPT );
#endif
      return;


    // Are we waiting for the ACK to confirm our "200 OK" message?
    case P2P_MSG_SESSION_OK:

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication: Got an ACK message (200 OK ACK)." << endl;
#endif

      // Reset, will be changed by initiateTransfer->contactStarted3..->sendDataPreparationAck.
      // TODO: rewrite the sending of the data-preparation message using the incoming/outgoing queue.
      waitingState_ = P2P_WAIT_DEFAULT;


      // This should happen after the "200 OK" message is sent.

      // If the user started the inviation,
      // we send the SLP OK ack, not receive it
      if( isUserStartedApp() )
      {
        // We received an 200 OK ACK but didn't expect it.
        kdWarning() << "P2PApplication: P2P message can't be handled, "
                    << "unexpected ACK response, waiting for 200 OK ACK (contact=" << getContactHandle() << ")." << endl;
        // TODO: cancel session in a different way, can't respond with ACKs on ACKs.
        shouldSendAck_ = true;  // ASSERT in sendP2PAck would fail otherwise
        sendP2PAck(P2PMessage::MSN_FLAG_ERROR);
        endApplication( i18n("The invitation was cancelled.  The contact sent bad data, or KMess doesn't support it.") );
        return;
      }

      // SLP OK message ACKed
      // For P2P apps, we can't prove any mime fields here, so an empty constructor is what's left..
      // This is where the data preparation ACK could be sent.
      contactStarted3_ContactConfirmsAccept(MimeMessage());

#ifdef KMESSTEST
      ASSERT( waitingState_ == P2P_WAIT_DEFAULT || waitingState_ == P2P_WAIT_FOR_PREPARE_ACK );
#endif

      // If the contactStarted3() didn't sent anything, we'll wait for incoming file data.
      if(waitingState_ == P2P_WAIT_DEFAULT)
      {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
        kdDebug() << "P2PApplication: Waiting state not changed, waiting for file data..." << endl;
#endif

        // For Pictures, the contact must send the data preparation ACK at this point (P2P_WAIT_FOR_PREPARE_ACK)
        // For Files, the contact must initiate the data transfer now. (P2P_WAIT_DEFAULT)
        waitingState_ = P2P_WAIT_FOR_FILE_DATA;
        waitingTimer_->start(30000, true); // 30 sec
      }

      return;


    case P2P_MSG_TRANSFER_OK:

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication: Got an ACK message (200 transfer OK ACK)." << endl;
#endif
      // For some reason, our "MSNSLP/1.0 200 OK" message can be ACKed when we just ACKed all data is received.
      // Detect this, otherwise the chat window is filled with "File transfer complete.. Contact aborted" messages.
      if( waitingState_ == P2P_WAIT_FOR_SLP_BYE )
      {
        kdWarning() << "P2PApplication: Received unexpected ACK for 'transrespbody' message "
                    << "(state=" << waitingState_ << " contact=" << getContactHandle() << ")." << endl;
        return;
      }

      waitingState_ = P2P_WAIT_DEFAULT;

      // This should happen after the "200 OK" message is sent.

      // SLP ransfer OK message ACKed
      if( ! applicationList_->hasDirectConnection()
      &&  ! applicationList_->hasPendingConnections() )
      {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
        kdDebug() << "P2PApplication: No connection attempts were made, so starting transfer immediately." << endl;
#endif
        initiateTransfer();
      }
      else
      {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
        kdDebug() << "P2PApplication: 200 transfer OK ACK received, but waiting for connection attempts to succeed or fail..." << endl;
#endif
      }

      // If the contactStarted3() didn't sent anything, we'll wait for incoming file data.
      if(waitingState_ == P2P_WAIT_DEFAULT)
      {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
        kdDebug() << "P2PApplication: Waiting state not changed, waiting for file data..." << endl;
#endif
        // For Files, the contact must initiate the data transfer now. (P2P_WAIT_DEFAULT)
        waitingState_ = P2P_WAIT_FOR_FILE_DATA;
        waitingTimer_->start(30000, true); // 30 sec
      }

      return;


    // Are we waiting for the ACK to confirm our data preparation message?
    case P2P_MSG_DATA_PREPARATION:
#ifdef KMESSTEST
      ASSERT( waitingState_ == P2P_WAIT_FOR_PREPARE_ACK );
      ASSERT( message.getAckDataSize() == 4 );
#endif

      waitingState_ = P2P_WAIT_DEFAULT;

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication: Got an ACK message (data-preparation ACK)." << endl;
#endif

      // This should happen after the data preparation message is sent.

      // If the user sent the invitation,
      // we shouldn't receive the prepare-ack, but sent it.
      if(isUserStartedApp())
      {
        // We received a data preparation ACK but didn't expect it.
        kdWarning() << "P2PApplication: P2P message can't be handled, "
                    << "unexpected ACK response, waiting for data-preparation-ACK (contact=" << getContactHandle() << ")." << endl;
        shouldSendAck_ = true;  // ASSERT in sendP2PAck would fail otherwise
        sendP2PAck(P2PMessage::MSN_FLAG_ERROR);
        endApplication( i18n("The invitation was cancelled.  The contact sent bad data, or KMess doesn't support it.") );
        return;
      }


      // Dispatch to derived implementation
      if( hasUnAckedMessage(P2P_MSG_TRANSFER_OK) )
      {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
        kdDebug() << "P2PApplication: Got an ACK message (data-preparation ACK, still waiting for 200 transfer OK ACK)." << endl;
#endif
        waitingState_ = P2P_WAIT_FOR_TRANSFER_ACK;
      }
      else
      {
        initiateTransfer();
      }
      return;


    // Are we waiting for the ACK to confirm that all data was received?
    case P2P_MSG_DATA:
#ifdef KMESSTES
      ASSERT( waitingState_ == P2P_WAIT_FOR_DATA_ACK );
#endif
      waitingState_ = P2P_WAIT_DEFAULT;

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication: Got an ACK message (all-received ACK)." << endl;
#endif

      // This should happen after all file/picture data was sent.


      if(isUserStartedApp())
      {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
        kdDebug() << "P2PApplication: All data sent and confirmed, sending BYE message..." << endl;
#endif

        // We sent the INVITE, now we send the BYE.
        waitingState_ = P2P_WAIT_DEFAULT;
        sendSlpBye();

        // All data was acked
        // Telling derived class the transfer is complete.
        showTransferComplete();

        // Don't terminate yet, that message needs to be ACK-ed as well.
      }
      else
      {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
        kdDebug() << "P2PApplication: All data sent and confirmed, waiting for BYE message..." << endl;
#endif

        // All data was acked, this is good enough.
        // The remaining teardown happens behind the scenes.
        showTransferComplete();

        // The other contact sent the INVITE, so we also wait for the BYE message.
        // This happens for picture transfers for example.

        waitingState_ = P2P_WAIT_FOR_SLP_BYE;
        waitingTimer_->start(30000, true); // 30 sec
      }

      return;



    // Are we waiting for the BYE ACK to quit the session/application?
    case P2P_MSG_SESSION_BYE:
#ifdef KMESSTEST
      ASSERT( waitingState_ == P2P_WAIT_FOR_SLP_BYE_ACK );
#endif

      waitingState_ = P2P_WAIT_DEFAULT;

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication: Got an ACK message (BYE ACK, terminating)." << endl;
#endif

      // Session complete.
      endApplication();
      return;


    // Are we expecting a ACK for our decline of the transfer?
    case P2P_MSG_TRANSFER_DECLINE:
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
        kdDebug() << "P2PApplication: Got an ACK message (transfer decline ACK)." << endl;
#endif
      return;


    // Are we waiting for the SLP Error message to be ACK-ed?
    case P2P_MSG_SLP_ERROR:
#ifdef KMESSTEST
      ASSERT( waitingState_ == P2P_WAIT_FOR_SLP_ERR_ACK );
#endif
      waitingState_ = P2P_WAIT_DEFAULT;

      if(! isUserStartedApp())
      {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
        kdDebug() << "P2PApplication: Got an ACK message (Confirmed SLP Error, waiting for BYE)." << endl;
#endif

        // We wait for the BYE
        waitingState_ = P2P_WAIT_FOR_SLP_BYE;
        waitingTimer_->start(30000, true); // 30 sec
      }
      else
      {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
        kdDebug() << "P2PApplication: Got an ACK message (Confirmed SLP Error, terminating)." << endl;
#endif
        // TODO: send BYE after the client confirmed our SLP error?
        endApplication();
      }

      return;


    // Standard ACK message or unhandled ACK message.
    default:


      // Avoid warnings when we're waiting for something else, and the ACK is not important.
      if( waitingState_ == P2P_WAIT_FOR_CONNECTION )
      {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
        kdDebug() << "P2PApplication: Got an ACK message." << endl;
#endif
      }
      else if( waitingState_ != P2P_WAIT_DEFAULT || messageType != P2P_MSG_UNKNOWN )
      {
        // Unexpected, this should be a normal ACK message.
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
        kdWarning() << "P2PApplication: Unhandled state for ACK message:"
                    << " waitingState=" << waitingState_ << " messageType=" << messageType << "!" << endl;
#endif
        // If we were waiting: the timer is stopped and
        // we forgot an if-block in this method.
        // So make sure the object will still be killed eventually.
        waitingTimer_->start(30000, true); // 30 sec
        return;
      }
      else
      {
        // Yes, this is truely a normal ack message
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
        kdDebug() << "P2PApplication: Got an ACK message." << endl;
#endif
      }
  }
}



// Got an direct connection handshake ack.
void P2PApplication::gotDirectConnectionHandshake(const P2PMessage &message)
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Got a direct connection handshake message." << endl;
#endif

  MsnDirectConnection *directConnection = applicationList_->getDirectConnection();

  // If the contact was sending a handshake when we dropped the connection,
  // it's possible this packet will be delivered over the switchboard.
  if( directConnection == 0 || ! directConnection->isConnected() )
  {
    // Ignore the packet
    return;
  }


  // Get connection info
  QString nonce    = message.getNonce();
  bool    isServer = directConnection->isServer();

  // Verify the nonce
  // Nonce is already used in ApplicationList to find this application,
  // but this is kept here for consistency, and if ApplicationList would change one time.
  if(nonce == nonce_)
  {
    // Mark connection as authorized
    directConnection->setAuthorized(true);

    // If we're the server, we still need to send a handshake back
    if(isServer)
    {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication: Contact sent correct nonce, sending direct connection handshake ACK." << endl;
#endif
#ifdef KMESSTEST
      ASSERT( waitingState_ == P2P_WAIT_FOR_HANDSHAKE );
#endif
      sendDirectConnectionHandshake();
      return;
    }
    else
    {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication: Contact sent correct nonce, marked connection as authorized." << endl;
#endif
#ifdef KMESSTEST
      ASSERT( waitingState_ == P2P_WAIT_FOR_HANDSHAKE_OK );
#endif
    }
  }
  else
  {
    // Invalid nonce! Drop connection
    kdWarning() << "P2PApplication: Contact sent invalid nonce (received " << nonce << " expected " << nonce_ << ")." << endl;
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
    kdDebug() << "P2PApplication: Closing connection, moving transfer to switchboard." << endl;
#endif

    // Like the official client, close the connection without any notice.
    // The transfer continues over the switchboard instead.
    directConnection->closeConnection();    // DirectConnectionPool deletes it.
  }
}



/**
 * Called when data is received
 *
 * @param  message  The P2P message containing the data.
 */
void P2PApplication::gotData(const P2PMessage &/*message*/)
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication - gotData" << endl;
#endif
}



/**
 * Called internally when data is received, this eventually calls gotData()
 */
void P2PApplication::gotDataFragment(const P2PMessage &message)
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Received file data, calling gotData()." << endl;
#endif
#ifdef KMESSDEBUG
  ASSERT( waitingState_ == P2P_WAIT_FOR_FILE_DATA );
#endif

  // TODO: do some verification on the size/offset fields??
  //       after all, the data can be sent over the public switchboard.

  uint dataMessageSize = message.getDataSize();
  uint dataOffset      = message.getDataOffset();


  // See if the class is trying to abort
  if( waitingState_ == P2P_WAIT_FOR_SLP_BYE_ACK )
  {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
    kdWarning() << "P2PApplication::gotDataFragment: Still receiving data while BYE is sent, sending error back." << endl;
#endif
    sendP2PAck(P2PMessage::MSN_FLAG_ERROR);
    return;
  }


  // Dispatch the message to the derived class
  showTransferProgress( dataOffset + dataMessageSize );
  gotData(message);


  // If we received the last part, ACK it and send the BYE
  if( ! message.isLastFragment() )
  {
    // Wait for more data to arrive.
    waitingTimer_->start(30000, true); // 30 sec
  }
  else
  {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
    kdDebug() << "P2PApplication::gotDataFragment: Got the last fragment, sending ACK." << endl;
#endif

    // All data received
    // TODO: For pictures, we can check whether this was really the picture data we expected.
    //       The file data needs to be hashed, and compared to the SHA1D field of the MsnObject.

    // Send the ACK
    sendP2PAck();

    // Signal derived class the transfer is complete
    showTransferComplete();

    // Determine what to send next
    if(isUserStartedApp())
    {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication::gotDataFragment: Application was started by the user, also send SLP BYE." << endl;
#endif

      // We sent the INVITE, now we send the BYE.
      waitingState_ = P2P_WAIT_DEFAULT;
      sendSlpBye();
      // Don't terminate yet, that message needs to be ACK-ed as well.
    }
    else
    {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication::gotDataFragment: Application was started by the contact, waiting for SLP BYE..." << endl;
#endif

      // Otherwise we wait for the BYE
      waitingState_ = P2P_WAIT_FOR_SLP_BYE;
      waitingTimer_->start(30000, true); // 30 sec
    }
  }
}



/**
 * Called when the data preparation message is received
 */
void P2PApplication::gotDataPreparation(const P2PMessage &/*message*/)
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Got the data preparation message" << endl;
#endif
#ifdef KMESSTEST
  ASSERT( shouldSendAck_ );
  ASSERT( waitingState_ == P2P_WAIT_FOR_PREPARE );
#endif


  // Tell the derived class it can send the data.
  // You typically want to call sendDataPreparation() or sendCancelMessage() here.
  // sendDataPreparationAck() only sets a boolean, so we're able to send the
  // correct ack or error message here.
  // The methods changes one the following variables:
  shouldSendAck_         = true;
  userShouldAcknowledge_ = true;

  initiateTransfer();


  // sendDataPreparation() was not called, determine now what to send:
  if(userShouldAcknowledge_)
  {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
    kdDebug() << "P2PApplication: Data preparation failed, sending reject" << endl;
#endif

    // Derived class still didn't acknowledge or cancel at all
    // (both disable userShouldAcknowledge_). This shouldn't occur!

    // Reject the data-preparation message
    sendP2PAck(P2PMessage::MSN_FLAG_ERROR);
    endApplication( i18n("The transfer failed.  Data preparation failed.") );
    return;
  }


  // If the user did acknowledge something, we no longer have to send an ack any more
  // it must have been sendCancelMessage(CANCEL_ABORT) calling sendP2PAck(P2PMessage::MSN_FLAG_ERROR).
  if(! shouldSendAck_)
  {
    // No error message here, assume your derived class already displayed something.
    // If sendP2PAck() was called, the app won't terminate, so end it now.
    endApplication();
    return;
  }


  // TODO: for WLM8, we need to delay the P2P ack message until the DC is established.

  // User did acknowledge, we can send the *actual* preparation ACK now.
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Data preparation successful, sending ACK" << endl;
#endif

  sendP2PAck();


  // Now wait for the data to arrive
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Waiting for file data to arrive..." << endl;
#endif

  waitingState_ = P2P_WAIT_FOR_FILE_DATA;
  waitingTimer_->start(30000, true); // 30 sec
}



/**
 * Received a P2P message with the error-flag set.
 * This method aborts the application.
 */
void P2PApplication::gotErrorAck(const P2PMessage &/*message*/)
{
  // Error in our headers?
  kdWarning() << "P2PApplication: Got a P2P error-response (flag=error contact=" << getContactHandle() << ")." << endl;

  if(dataType_ == P2P_TYPE_NEGOTIATION)
  {
    // We likely made an error in the invitation/bye reponse
    contactAborted( /* defaults to getContactAbortMessage() */ );
  }
  else
  {
    // TODO: make this error custom.
    // We were sending or receiving a file/picture
    // Can also happen if the contact can't handle our data for a different reason.
    contactAborted( i18n("The transfer failed.") );
  }
}



/**
 * Parse an incoming message
 *
 * @param  message  A reference to a P2PMessage object.
 */
void P2PApplication::gotMessage(const P2PMessage &p2pMessage)
{
  // Reset session variables.
  waitingTimer_->stop();   // stop timer
  gotSlpMessage_ = false;  // tell this is not a slp message until proven otherwise

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Message size=" << p2pMessage.getDataSize()
                           << " total="        << p2pMessage.getTotalSize()
                           << " offset="       << p2pMessage.getDataOffset() << endl;
#endif


  // If this object was initialized to handle a Bad packet.
  if(getMode() == APP_MODE_ERROR_HANDLER)
  {
    // Make an exception for SLP packets
    bool isSlpMessage = (p2pMessage.getSessionID() == 0 &&
                         p2pMessage.getFlags()     == 0);
    if(! isSlpMessage)
    {
      // We received an unknown message, this object was initialized to send the correct response.
      shouldSendAck_ = true;  // ASSERT in sendP2PAck would fail otherwise
      sendP2PAck(P2PMessage::MSN_FLAG_ERROR);
      endApplication( i18n("The invitation was cancelled.  The contact sent bad data, or KMess doesn't support it.") );
      return;
    }
  }


  // Handle ACKs before everything else.
  if( p2pMessage.getDataSize() == 0 )  // HACK: for Bot2K3 4.1: no testing for flags
  {
    UnAckedMessage *unAcked    = getUnAckedMessage(p2pMessage);
    P2PMessageType messageType = P2P_MSG_UNKNOWN;
    if( unAcked == 0 )
    {
      if( ! p2pMessage.isConnectionHandshake() ) // exception, no original message set.
      {
        kdWarning() << "P2PApplication: Unable to handle ACK message, original message not found"
                    << " (flags=0x" << QString::number(p2pMessage.getFlags(), 16)
                    << " size=0 state=" << waitingState_ << " contact=" << getContactHandle() << ")." << endl;
        return;
      }
    }
    else
    {
      // Remove record, auto-deletes
      // Avoids that hasUnAckedMessage() returns the message we just received.
      messageType = unAcked->messageType;
      outgoingMessages_.remove(unAcked);
    }


    if( p2pMessage.isAck() )
    {
      gotAck( p2pMessage, messageType );
    }
    else if(p2pMessage.isConnectionHandshake())
    {
      gotDirectConnectionHandshake( p2pMessage );
    }
    else if(p2pMessage.isError())
    {
      gotErrorAck( p2pMessage );
    }
    else if(p2pMessage.isWaiting())
    {
      gotTimeoutAck( p2pMessage );
    }

    return;
  }


  // Send the ack if needed (also applies to INVITE messages)
  shouldSendAck_ = p2pMessage.isLastFragment();

  // Store some message parameters for Acknowledgement later, or to send an error message.
  // This should not be inserted in a if(shouldSendAck_) line, this info is used to resolve fragmented messages.
  lastIncomingMessage_.dataSize    = p2pMessage.getDataSize();
  lastIncomingMessage_.messageID   = p2pMessage.getMessageID();
  lastIncomingMessage_.messageType = P2P_MSG_UNKNOWN;
  lastIncomingMessage_.sentTime    = 0;
  lastIncomingMessage_.sessionID   = p2pMessage.getSessionID();
  lastIncomingMessage_.totalSize   = p2pMessage.getTotalSize();
  lastIncomingMessage_.uniqueID    = p2pMessage.getAckMessageID();


  // Session ID is 0? -> the clients negotiate for session.
  if(p2pMessage.getSessionID() == 0)
  {
    if(p2pMessage.getFlags() == 0)
    {
      // If this object was initialized to handle a Bad packet.
      if(getMode() == APP_MODE_ERROR_HANDLER)
      {
        // This object was initialized to send an error response for an unknown packet (at SLP level).
        showMessage( i18n("The transfer failed.  The contact sent bad data, or KMess doesn't support it.") );
        sendP2PAck();
        sendSlpError("500 Internal Error"); // Waits for ACK and terminates
        return;
      }


      // Parse the fragment
      gotNegotiationFragment(p2pMessage);
    }
    else
    {
      kdWarning() << "P2PApplication: P2P message can't be handled, negotiation message type "
                  << "(flags=0x" << QString::number(p2pMessage.getFlags(), 16)
                  << " size="    << p2pMessage.getDataSize()
                  << " contact=" << getContactHandle() << ")." << endl;

      // Don't know what to do with a negotiation message with flags set.
      sendP2PAck(P2PMessage::MSN_FLAG_ERROR);
      endApplication( i18n("The transfer failed.  The contact sent bad data, or KMess doesn't support it.") );
      return;
    }
  }
  // Session ID not zero? -> clients exchange data in the session
  else
  {
    // Check for data preparation messages
    if( p2pMessage.getFlags()     == 0
    &&  p2pMessage.getDataSize()  == 4
    &&  p2pMessage.getTotalSize() == 4 )
    {
      gotDataPreparation(p2pMessage);
    }
    else if( waitingState_ != P2P_WAIT_FOR_FILE_DATA
         &&  p2pMessage.isPictureData()
         &&  p2pMessage.getDataSize()  == 4
         &&  p2pMessage.getTotalSize() == 4 )
    {
      // HACK: added for Encarta Instant Answers (has data flag set with preparation message).
      kdWarning() << "P2PApplication: Expecting data-preparation message, "
                  << "got message of 4 bytes with incorrect flags "
                  << "(assuming data preparation, contact=" << getContactHandle() << ")!" << endl;
      gotDataPreparation(p2pMessage);
    }
    // Check for data messages
    else if(p2pMessage.isPictureData()
         || p2pMessage.isFileData())
    {
      gotDataFragment(p2pMessage);
    }
    else if(waitingState_ == P2P_WAIT_FOR_FILE_DATA
         && p2pMessage.getFlags() == 0
         && p2pMessage.isFragment())
    {
      // HACK: added for Kopete 0.9.2 code (has no flag set with data messages).
      kdWarning() << "P2PApplication: Expecting data message, "
                  << "got fragmented message with no flags set "
                  << "(assuming file data, contact=" << getContactHandle() << ")!" << endl;
      gotDataFragment(p2pMessage);
    }
    else
    {
      // Unknown p2p message
      kdWarning() << "P2PApplication: P2P message can't be handled, unknown data message type "
                  << "(flags=0x" << QString::number(p2pMessage.getFlags(), 16)
                  << " size="    << p2pMessage.getDataSize()
                  << " contact=" << getContactHandle() << ")." << endl;

      // Got an P2P binary message we can't handle. (some new kind of flag..?)
      shouldSendAck_ = true;  // ASSERT in sendP2PAck would fail otherwise
      sendP2PAck(P2PMessage::MSN_FLAG_ERROR);
      endApplication( i18n("The transfer failed.  The contact sent bad data, or KMess doesn't support it.") );
      return;
    }
  }

#ifdef KMESSTEST
  // At the end of this method, we should have sent the ACK..!
  ASSERT( ! shouldSendAck_ );
#endif
}



/**
 * Got a message fragment with SessionID 0
 * All fragments are buffered and
 * gotNegotiationMessage() is called when all parts are received
 *
 * @param  p2pMessage  The received P2P message
 */
void P2PApplication::gotNegotiationFragment(const P2PMessage &p2pMessage)
{
  if(! p2pMessage.isLastFragment())
  {
    // Not everything received yet, buffer it.
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
    kdDebug() << "P2PApplication: Got an SLP fragment, buffering it." << endl;
#endif

    if(buffer_ == 0)
    {
      buffer_ = new QBuffer();
      buffer_->open(IO_ReadWrite);
    }

    buffer_->writeBlock( p2pMessage.getData(), p2pMessage.getDataSize() );
  }
  else
  {
    // Read the message or buffer
    QString slpMessage;
    QByteArray bufferData;

    if(buffer_ != 0)
    {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication: Got the last SLP fragment." << endl;
#endif

      // Append the last MSNSLP fraqment and call the method.
      buffer_->writeBlock( p2pMessage.getData(), p2pMessage.getDataSize() );
      buffer_->reset();

      // Read the buffer
      bufferData = buffer_->readAll();
      slpMessage = QString::fromUtf8( bufferData.data(), bufferData.size() );

      delete buffer_;
      buffer_ = 0;
    }
    else
    {
      // Read the data directly from the message
      slpMessage = QString::fromUtf8( p2pMessage.getData(), p2pMessage.getDataSize() );
    }


    // Parse the negotiation message
    gotNegotiationMessage( slpMessage );
  }
}



/**
 * Parse SLP "session negotiation messages".
 * Those messages don't have a sessionID set, and have
 * MSNSLP Mime fields in the body.
 *
 * @param  slpMessage  The MSNSLP message to parse.
 */
void P2PApplication::gotNegotiationMessage(const QString &slpMessage)
{
  // Parse the "negotiation message" (P2PMessage with sessionID 0)

  // Parse the MIME-fields (skip the first line, because it's not in MIME format)
  int      preambleEnd = slpMessage.find("\r\n");
  QString  preamble    = slpMessage.left( (preambleEnd == -1) ? 20 : preambleEnd     );
  QString  mimeFields  = slpMessage.mid(  (preambleEnd == -1) ?  0 : preambleEnd + 2 );
  MimeMessage slpMimeMessage( mimeFields );


  // Verify it is indeed directed to us..
  QString msgTo = slpMimeMessage.getValue("To");

  if(! msgTo.isEmpty())
  {
    // Only test if the field is available. If it's not, either something
    // is wrong in this application, or the client sent a bad message and
    // we should respond with an 500 error instead.
    if( msgTo != "<msnmsgr:" + CurrentAccount::instance()->getHandle() + ">" )
    {
      if(shouldSendAck_) sendP2PAck();

      // In the future, the <msnmsgr: > string might be replaced with
      // something to support MSN to AOL chats, etc..
      kdWarning() << "P2PApplication: P2P message can't be handled, "
                  << "addressed to someone else (contact=" << getContactHandle() << ")." << endl;

      showMessage( i18n("The invitation was cancelled.  Message was not directed to us.") );
      sendSlpError("404 Not Found"); // Waits for ACK and terminates
      return;
    }
  }

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Got an SLP message, preamble=" << preamble << "." << endl;
#endif

  // Find out what kind of message this is:
  if(preamble.startsWith("INVITE MSNMSGR"))
  {
    // Got an invitation!
    gotSlpInvite(slpMimeMessage);
  }
  else if(preamble.startsWith("MSNSLP/"))
  {
    // Error message, or "MSNSLP/1.0 200 OK" message
    gotSlpStatus(slpMimeMessage, preamble);
  }
  else if(preamble.startsWith("BYE MSNMSGR"))
  {
    // Got a BYE message
    gotSlpBye(slpMimeMessage);
  }
  else
  {
    // Unknown SLP message
    kdWarning() << "P2PApplication: P2P message can't be handled, unsuppored SLP negotiation message "
                << "(preamble=" << preamble << " contact=" << getContactHandle() << ")." << endl;

    // Got an unknown SLP preamble
    sendP2PAck(P2PMessage::MSN_FLAG_ERROR);
    endApplication( i18n("The invitation was cancelled.  The contact sent bad data, or KMess doesn't support it.") );
    return;
  }
}



/**
 * Got an MSNSLP INVITE message
 */
void P2PApplication::gotSlpBye(const MimeMessage &/*slpMimeMessage*/)
{
  // If the BYE was received early, it means the session was aborted prematurely.
  //
  // There appears to be a small bug in MSN 7.0. It doesn't sent the final "all data received" ACK,
  // but sends the BYE directly. After we ACK it, we get a "BYE sent" message back.
  // This issue appears to be fixed in MSN 7.5.


  if(shouldSendAck_)
  {
    sendP2PAck();
  }

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Got SLP BYE message, closing application." << endl;
#endif

  if( waitingState_ != P2P_WAIT_FOR_SLP_BYE && waitingState_ != P2P_WAIT_FOR_DATA_ACK )
  {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
    kdDebug() << "P2PApplication: BYE message was unexpected; contact aborted "
              << "(state=" << waitingState_ << " contact=" << getContactHandle() << ")" << endl;
#endif

    // End with failure, bye sent at a different moment.
    contactAborted();   // closes the application

    // TODO: send a 0x40 or 0x80 packet here.
    // When we started the invitation, send a 0x40 packet to indicate we closed our invite early.
    // When the contact started the invite, send a 0x80 packet to incide we closed their invite early.
  }
  else
  {
    // Normal end
    endApplication();
  }
}



/**
 * Got an MSNSLP INVITE message
 */
void P2PApplication::gotSlpInvite(const MimeMessage &slpMimeMessage)
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Got SLP INVITE message" << endl;
#endif
#ifdef KMESSTEST
  ASSERT( waitingState_ == P2P_WAIT_DEFAULT        // First invitation
       || waitingState_ == P2P_WAIT_FOR_FILE_DATA  // Second invitation for file transfers.
       || waitingState_ == P2P_WAIT_FOR_PREPARE    // WLM8: second invitation for picture transfer
        );
#endif

  // Reset the waiting state if this is the second INVITE for a file transfer
  // TODO: test what to do for the new WLM8 feature (sending transfer invites before data preparation).
  if(waitingState_ == P2P_WAIT_FOR_FILE_DATA)
  {
    waitingState_ = P2P_WAIT_DEFAULT;
  }

  // Session invitation

  // Extract the fields of the message. This is required for userRejected()
  MimeMessage slpMimeContent(slpMimeMessage.getBody());

  // Set global fields from INVITE message.
  QString slpVia         = slpMimeMessage.getValue("Via");
  callID_                = slpMimeMessage.getValue("Call-ID");
  invitationCSeq_        = slpMimeMessage.getValue("CSeq").toInt();
  invitationContentType_ = slpMimeMessage.getValue("Content-Type");

  // Extract branch from the "Via" parameter
  QRegExp callRE(";branch=(.+)");     // don't use a guid-pattern here, msn6 seams to accept random strings.
  callRE.search(slpVia);
  branch_ = callRE.cap(1);


  // Don't forget the initialize the base class
  if( getCookie().isEmpty() )
  {
    startByInvite(generateCookie());
  }


  // Indicate this is an SLP message (sendCancelMessage() uses this)
  // This value is reset once a message is sent.
  gotSlpMessage_ = true;


  // Extract the body of the SLP message
  MimeMessage slpContent(slpMimeMessage.getBody());

  // Handle invitation based on the Content-Type field
  if( invitationContentType_ == "application/x-msnmsgr-sessionreqbody" )
  {
    // Client invites us for a session
    // Read requested session id from the message
    invitationSessionID_ = slpMimeContent.getValue("SessionID").toULong();

    // Don't accept invites if we sent one.
    if(messageID_ != 0 && isUserStartedApp())
    {
      kdWarning() << "P2PApplication: Got an INVITE response, rejecting (contact=" << getContactHandle() << ")." << endl;

      unsigned long tempSessionID = sessionID_;
      sessionID_ = 0;
      sendP2PAck(P2PMessage::MSN_FLAG_ERROR);
      sessionID_ = tempSessionID;
      return;
    }

    // Send the ACK message
    if(shouldSendAck_)
    {
      sendP2PAck();
    }

    // Tell the derived class we've got an invitation
    contactStarted1_ContactInvitesUser(slpContent);
  }
  else if( invitationContentType_ == "application/x-msnmsgr-transreqbody")
  {
    // Send the ACK message
    if(shouldSendAck_)
    {
      sendP2PAck();
    }

    // Client requested to transfer the session
    // Handle the invitation internally
    gotSlpTransferInvitation( slpMimeContent );
  }
  else
  {
    // Send the ACK message
    if(shouldSendAck_)
    {
      sendP2PAck();
    }

    // Client sent an unknown invitation type
    kdWarning() << "P2PApplication: Received unexpected Content-Type "
                << "(type=" << invitationContentType_ << " contact=" << getContactHandle() << ")." << endl;

    // Indicate we don't like that content-type:
    sendCancelMessage(CANCEL_INVALID_SLP_CONTENT_TYPE);
    // Don't QUIT, the error will be ACK-ed.
    return;
  }
}



/**
 * Got an MSNSLP Status header
 */
void P2PApplication::gotSlpStatus(const MimeMessage &slpMimeMessage, const QString &preamble)
{
  // An Error message, or "MSNSLP/1.0 200 OK" message
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Got SLP status message (" << preamble << ")" << endl;
#endif


  // Should only be received if we started it.
  if(! isUserStartedApp())
  {
    kdWarning() << "P2PApplication: P2P message can't be handled, unexpected SLP status response "
                << "(preamble=" << preamble << " contact=" << getContactHandle() << ")." << endl;

    sessionID_ = 0;
    sendP2PAck(P2PMessage::MSN_FLAG_ERROR);
    endApplication( i18n("The invitation was cancelled.  The contact sent bad data, or KMess doesn't support it.") );
    return;
  }


  // Parse the preamble
  QRegExp re("MSNSLP/\\d+\\.\\d+ (\\d+) ([^\r\n]+)");
  re.search(preamble);
  int     code   = re.cap(1).toUInt();
  QString detail = re.cap(2);


  // Reset the waiting state if we were waiting.
  if(waitingState_ == P2P_WAIT_CONTACT_ACCEPT)
  {
    waitingState_ = P2P_WAIT_DEFAULT;
  }

  // Handle the status codes


  // If the invitation was accepted
  if(code == 200)
  {
    // We don't need to do much here:
    // The message will be ACK-ed automatically with the correct session ID
    // because that ID was given in the sendSlpInvitation() method.

    // This the first ACK to be sent with the SessionID set.
    if(shouldSendAck_)
    {
      sendP2PAck();
    }

    // With a 200 code, we reset the waiting state for every situation,
    // to make the next check work.
    waitingState_ = P2P_WAIT_DEFAULT;

    // Parse the data of the message
    QString     contentType = slpMimeMessage.getValue("Content-Type");
    MimeMessage slpContent(slpMimeMessage.getBody());  // body consists of more mime fields

    // The content-type tells us which fields are present
    if( contentType == "application/x-msnmsgr-sessionreqbody" )
    {
      // Message confirms the SessionID
#ifdef KMESSTEST
      ASSERT( slpContent.getValue("SessionID").toULong() == getSessionID() );
#endif

      // Tell the derived class the invitation was accepted
      userStarted2_ContactAccepts(slpContent);
    }
    else if( (contentType == "application/x-msnmsgr-transrespbody") ||
              // HACK: allow transreqbody too for broken clients, including KMess 1.4.2:
             (contentType == "application/x-msnmsgr-transreqbody" && slpContent.hasField("Nonce") )) 
    {
      // Message tells us how we can create the direct connection.
      // This response is handled here, so it's transparent for the derived class.
      gotSlpTransferResponse(slpContent);
    }
    else
    {
      // Unsupported response type.
      kdWarning() << "P2PApplication: Received unexpected 200/OK message "
                  << "(type=" << contentType << " contact=" << getContactHandle() << ")!" << endl;
      showMessage( i18n("The invitation was cancelled.  The contact sent bad data, or KMess doesn't support it.") );
      sendCancelMessage( CANCEL_INVALID_SLP_CONTENT_TYPE );
      return;
    }

    // Determine whether we should be waiting for the data-preparation message.
    // If we didn't send anything that changed the state and no transfer has started,
    // we're likely waiting for some kind of prepare message
    if(waitingState_ == P2P_WAIT_DEFAULT && dataType_ == P2P_TYPE_NEGOTIATION)
    {
      // Waiting for prepare message
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication: Waiting for contact to send some prepare message..." << endl;
#endif

      waitingState_ = P2P_WAIT_FOR_PREPARE;
      waitingTimer_->start(30000, true); // 30 sec
    }

    return;
  }


  // We sent a bad address in the to: header
  if(code == 404)
  {
    if(shouldSendAck_)
    {
      sendP2PAck();
    }

    // Inform the user KMess has an error
    kdWarning() << "P2PApplication: KMess sent a bad address in the 'to:' header "
                << "(contact=" << getContactHandle() << ")." << endl;
    contactAborted( i18n("The contact rejected the invitation.  An internal error occured.") );
    return;
  }


  // Something was not supported
  if(code == 500)
  {
    if(shouldSendAck_)
    {
      sendP2PAck();
    }

    if(slpMimeMessage.getValue("Content-Type") == "null")
    {
      // The Content-Type was not supported
      kdWarning() << "P2PApplication: Content-Type of our invitation was not understood "
                  << " (type=" << invitationContentType_ << " contact=" << getContactHandle() << ")!" << endl;
    }
    else
    {
      // Bad header sent in SLP fields, or invitation type was not supported
      kdWarning() << "P2PApplication: Our invitation was not supported or the contact client had an internal error "
                  << "(contact=" << getContactHandle() << ")" << endl;
    }

    contactAborted( i18n("The contact rejected the invitation.  An internal error occured.") );
    return;
  }


  // The invitation was declined
  if(code == 603)
  {
    if(shouldSendAck_)
    {
      sendP2PAck();
    }

    // Contact Declined
    contactRejected();
    return;
  }


  // Last option: tell the contact it sent an unknown message

  kdWarning() << "P2PApplication: P2P message can't be handled, unsupported SLP response "
              << "(preamble=" << preamble << " contact=" << getContactHandle() << ")." << endl;

  // Got an unknown SLP status code
  sendP2PAck(P2PMessage::MSN_FLAG_ERROR);
  endApplication( i18n("The invitation was cancelled.  The contact sent bad data, or KMess doesn't support it.") );
}



/**
 * Got an MSNSLP transfer invitation.
 * This is typically received when the contact initiated the file transfer session.
 * The message contains meta-information about the contact's network configuration.
 * In the response message, our IP address is returned back.
 */
void P2PApplication::gotSlpTransferInvitation(const MimeMessage &slpContent)
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Invite message is a request to transfer the session." << endl;
#endif

  // For some reason it's also possible to receive SLP transfer invitations
  // with an entirely different callid while trying to setup a direct connection.
  if( sessionID_ == 0 )
  {
    kdWarning() << "P2PApplication: Received an invitation for a direct connection,"
                << " but no session is set up (maybe a msn7 feature, contact=" << getContactHandle() << ")." << endl;
  }

  if(applicationList_->hasDirectConnection() )
  {
    kdWarning() << "P2PApplication: Received an invitation to start a direct connection, "
                << "but a direct connection is already available "
                << "(rejecting invitation, "
                << " authorized=" << applicationList_->getDirectConnection()->isAuthorized()
                << " contact=" << getContactHandle() << ")." << endl;

    // Decline the invitation forcefully with an error.
    // Other client should still continue with the file transfer
    // TODO: test the effect of this.
    sendCancelMessage(CANCEL_ABORT);
    return;
  }

  // TODO: Quick fix for WLM8: don't allow direct connection invite for picture transfer
  if( isUserStartedApp() )
  {
    waitingState_ = P2P_WAIT_DEFAULT;  // avoid assertion.
    sendSlpError("603 Decline", invitationSessionID_, invitationContentType_, P2P_MSG_TRANSFER_DECLINE);
    return;
  }

  // Create the accept message
  MimeMessage acceptMessage;
  bool startListening = false;

  // Get network details
  QString localIp    = getLocalIp();
  QString externalIp = getExternalIp();

  // Read the data from the message
  QString bridges        = slpContent.getValue("Bridges");       // Transport layers the client supports
  int     netID          = slpContent.getValue("NetID").toInt(); // Some ID
  QString connectionType = slpContent.getValue("Conn-Type");     // Suggested connection type
  QString hasUPnPNat     = slpContent.getValue("UPnPNat");       // uses UPnP-enabled NAT router
  QString hasICF         = slpContent.getValue("ICF");           // uses Internet Connection Firewall


  // Based on the contact's network configuration,
  // we choose how we want to send the file.
  //
  // The client can include a Hashed-Nonce field,
  // and we can respond with another Hashed-Nonce field.
  // This algorithm is not known, but including a normal Nonce field
  // in the accept message seams to work (the client has some fallback).

  // Listen at the internal port
  // The connection can be used for both the internal and external port
  // because it only accepts one incoming packet, the second one will be rejected.
  int listeningPort = applicationList_->addServerConnection();

  if( listeningPort != 0 )
  {
    // Get parameters for message
    nonce_ = generateGUID();  // used for authentication

    acceptMessage.addField( "Bridge",    "TCPv1" );
    acceptMessage.addField( "Listening", "true"  );
    acceptMessage.addField( "Nonce",     nonce_  );

    if(localIp != externalIp)
    {
      acceptMessage.addField("IPv4Internal-Addrs", localIp    );
      acceptMessage.addField("IPv4Internal-Port",  QString::number(listeningPort) );
    }

    acceptMessage.addField("IPv4External-Addrs", externalIp );   // TODO: include other aliases
    acceptMessage.addField("IPv4External-Port",  QString::number(listeningPort) );

    // Inform the user about the file transfer progress.
    showTransferMessage( i18n("Awaiting connection at %1, port %2").arg(externalIp).arg(listeningPort) );

    // Prepare to send the nonce too when connection to other contact is established
    connect( applicationList_, SIGNAL(     connectionEstablished() ),   // when connected, send nonce
             this,             SLOT  ( slotConnectionEstablished() ));
    connect( applicationList_, SIGNAL(      connectionAuthorized() ),   // when authorized, send data
             this,             SLOT  (  slotConnectionAuthorized() ));
    connect( applicationList_, SIGNAL(          connectionFailed() ),   // when failed, revert to switchboard
             this,             SLOT  (      slotConnectionFailed() ));

    startListening = true;  // avoids double-checking certain conditions.
  }
  else
  {
    // Could not listen at a local port, use switchboard instead.
    kdWarning() << "P2PApplication: Could not listen at local port, redirecting transfer to the switchboard." << endl;

    // This configuration sends each file over the switchboard:
    acceptMessage.addField( "Bridge",    "TCPv1" );
    acceptMessage.addField( "Listening", "false" );
    acceptMessage.addField( "Nonce",     "{00000000-0000-0000-0000-000000000000}");

    // Inform the user about the file transfer progress.
    showTransferMessage( i18n("Reverting to indirect file transfer (this could be slow).") );
    // TODO: at this point, it's assumend the other client will initiate the transfer.
    //       perhaps set the waitingState_ to detect when the contact does not return a request?
  }

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Sending 200 transfer OK message, listening=" << (listeningPort != 0) << "." << endl;
#endif

  // Confirm the session
  sendSlpOkMessage( acceptMessage );

  // Wait for the contact to ACK the message..
  if( startListening )
  {
    waitingState_ = P2P_WAIT_FOR_CONNECTION;
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
    kdDebug() << "P2PApplication: Waiting for a connection to succeed or fail..." << endl;
#endif
  }
  else
  {
    // Expecting file data now we've rejected the invite.
    // TODO: it's also possible another invite is sent, hoping for a second chance.
    waitingState_ = P2P_WAIT_FOR_FILE_DATA;
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
    kdDebug() << "P2PApplication: Waiting for contact to send file data..." << endl;
#endif
  }
  waitingTimer_->start(30000);  // 30 secs.
}



/**
 * Got an MSNSLP transfer invitation response.
 * This is received when we initiate a connection invitation with sendSlpTransferInvitation().
 * It's a MSNSLP/200 OK message, the body optionally contains the IP address to connect to.
 */
void P2PApplication::gotSlpTransferResponse(const MimeMessage &slpContent)
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication::gotSlpTransferResponse: Parsing 200 OK message." << endl;
#endif

  // Test for an existing (possibly even authorized) direct connection.
  // This could happen with local-lan file transfer, the switchboard roundtrip for the response message could take longer.
  if( applicationList_->hasDirectConnection() )
  {
    if( applicationList_->getDirectConnection()->isAuthorized() )
    {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication::gotSlpTransferResponse: Connection already authorized "
                << "before invitation response was received, starting transfer." << endl;
#endif
      initiateTransfer();
    }
    else
    {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication::gotSlpTransferResponse: Connection already established "
                << "before invitation response was received, waiting for connection to authorize..." << endl;
#endif
      // Slots already connected at this point.
    }
    return;
  }


  // The Conn-type field of the message is the least important here,
  // the "Listening", "ICF" and "UPnPNat" tell us more about the situation of the contact,
  // and the probability to create a successful direct connection in either way.

  // Get fields from message
  bool listening      = slpContent.getValue("Listening") == "true";
  bool listenInternal = slpContent.hasField("IPv4Internal-Addrs");
  bool listenExternal = slpContent.hasField("IPv4External-Addrs");
  nonce_              = slpContent.getValue("Nonce");

  // If the contact is not listening, we can do the following things:
  // - wait until the contact connects to a port we're listening on (not implemented yet)
  // - send the data over the switchboard instead. (always works, but it's slow).

  // If the contact is listening, try to create a direct connection to the contact.
  if( listening && (listenExternal || listenInternal) )
  {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
    kdDebug() << "P2PApplication::gotSlpTransferResponse: Contact is listening, attempting to connect to the given IP-address." << endl;
#endif

    QStringList ipAddresses;
    uint        port;

    // Connect to the internal IP address
    if( listenInternal )
    {
      // Create a direct connection for the internal IP-address.
      ipAddresses = QStringList::split( " ", slpContent.getValue("IPv4Internal-Addrs") );
      port        = slpContent.getValue("IPv4Internal-Port").toUInt();
      for ( QStringList::Iterator it = ipAddresses.begin(); it != ipAddresses.end(); ++it )
      {
        if( ! applicationList_->hasDirectConnection()
        && applicationList_->addConnection( *it, port ) )
        {
          showTransferMessage( i18n("Connecting to %1, port %2").arg(*it).arg(port) );
        }
      }
    }

    // Connect to the external IP address (unless a connection was made really quick)
    if( listenExternal && ! applicationList_->hasDirectConnection() )
    {
      // Create a direct connection for the external IP-address.
      ipAddresses  = QStringList::split( " ", slpContent.getValue("IPv4External-Addrs") );
      bool sameLan = ( listenInternal && ipAddresses.contains( CurrentAccount::instance()->getExternalIp() ) );
      port         = slpContent.getValue("IPv4External-Port").toUInt();
      for ( QStringList::Iterator it = ipAddresses.begin(); it != ipAddresses.end(); ++it )
      {
        if( ! applicationList_->hasDirectConnection()
        && applicationList_->addConnection( *it, port )
        && ! sameLan )  // hide "connecting to <external>" message when hosts are _likely_ at the same lan.
        {
          showTransferMessage( i18n("Connecting to %1, port %2").arg(*it).arg(port) );
        }
      }
    }
  }

  // After passing the connections to addConnection() it's possible
  // some slots are already fired and the connection was established.

  // See if there aren't connections established already.
  if( ! applicationList_->hasDirectConnection() )
  {
    if( applicationList_->hasPendingConnections() )
    {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication::gotSlpTransferResponse: Waiting for connection attempt to succeed..." << endl;
#endif

      // Wait for connection to establish
      waitingState_ = P2P_WAIT_FOR_CONNECTION2;
      waitingTimer_->start(30000, true); // 30 sec
    }
    else
    {
      // Nothing is pending, nothing is connected.
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication::gotSlpTransferResponse: No connection possible, informing ApplicationList." << endl;
#endif

      // If no direct connection attempt will be made.
      // Reset the pending invitation connection flag, ApplicationList sends the connectionFailed signal.
      // This also signals other applications the connection attempt failed.
      applicationList_->setPendingConnectionInvitation(false);
    }
  }
}



/**
 * Received a message with the timeout-flag set.
 * This method aborts the application.
 */
void P2PApplication::gotTimeoutAck(const P2PMessage &/*message*/)
{
  // Error in our headers?
  if(isWaitingForUser())
  {
    // Contact aborted this session because we didn't press "accept/cancel" within 30 secs.
    contactAborted( i18n("The transfer failed.  Timeout while waiting for user to accept.") );
  }
  else
  {
    // Apparently the contact was waiting for something, likely some packet we had to send.
    kdDebug() << "P2PApplication: WARNING - Got a P2P error-response (flag=waiting"
              << " localstate=" << waitingState_ << " contact=" << getContactHandle() << ")." << endl;
    contactAborted( i18n("The transfer failed.  An internal error occured.") );
  }
}



/**
 * Verify whether the application can handle the given ACK message.
 */
bool P2PApplication::hasUnAckedMessage(const P2PMessage &p2pMessage)
{
  // Simply test using the internal function, don't expose the UnAckedMessage pointer to external classes.
  return (getUnAckedMessage(p2pMessage) != 0);
}



/**
 * Verify whether a given message is still unacked.
 */
bool P2PApplication::hasUnAckedMessage(const P2PMessageType messageType)
{
  QPtrListIterator<UnAckedMessage> it( outgoingMessages_ );
  while( it.current() != 0 )
  {
    if( it.current()->messageType == messageType )
    {
      return true;
    }
  }

  return false;
}



/**
 * Signal the derived class it can initiate the file transfer.
 */
void P2PApplication::initiateTransfer()
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Signalling implementation class to start the file transfer." << endl;
#endif

  if( waitingState_ == P2P_WAIT_FOR_CONNECTION
  ||  waitingState_ == P2P_WAIT_FOR_CONNECTION2 )
  {
    waitingState_ = P2P_WAIT_DEFAULT;
  }

  // When we start sending information, it could take a while before
  // the contact returns messages which disable the waiting timer.
  // Avoid this by disabling it now, before we kill our own transfer.
  if( waitingTimer_->isActive() )
  {
    kdWarning() << "P2PApplication::initiateTransfer: timeout detection timer is still running "
                << "(state=" << waitingState_ << ")." << endl;
    waitingTimer_->stop();
  }


  // Call the connect implementation based on who invoked this application.
  // The implementations of these methods can call sendData(), and
  // the actual transfer will be handled transparently here.
  if( isUserStartedApp() )
  {
    userStarted3_UserPrepares();
  }
  else
  {
    contactStarted4_ContactConfirmsPreparation();
  }
}



/**
 * The switchboard requested to resume the transfer.
 */
void P2PApplication::resumeTransfer()
{
  if( ! isTransferPaused() )
  {
    return;
  }

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Application is allowed to resume transfer." << endl;
#endif

  // Call parent, reset isTransferPaused()
  Application::resumeTransfer();

  if( dataSource_ != 0 && dataType_ != P2P_TYPE_NEGOTIATION)
  {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
    kdDebug() << "P2PApplication: Data source was set, rescheduling transfer slot." << endl;
#endif

    QTimer::singleShot( 10, this, SLOT(slotSendData()) );
  }
}



/**
 * Send a cancel message. Some errors also terminate the application eventually.
 *
 * @param  cancelReason  The reason why the application is cancelled.
 *
 * MSNSLP related notes:
 *   Error reasons sending an MSNSLP error messages should only be
 *   used in the contactStarted1 / contactStarted2 methods..
 *   Once the MSNSLP error message is sent, P2PApplication waits for the
 *   contact to ACK, and terminates the application afterwards.
 *   Hence, you don't have to call endApplication() from contactStarted1 / contactStarted2.
 *
 *   When an MSNSLP error is sent at a different moment, this will likely result in
 *   undefined responses from the other client (or the error message might be ignored).
 *   If KMess is compiled in debug mode, an assertion in sendSlpError() will detect this.
 *
 *
 * The following cancelReason values are supported:
 * - CANCEL_INVITATION
 *     The user declined the invitation.
 *
 *     This sends an "MSNSLP 603 Decline" message, waits for
 *     the contact to ACK and terminates this application afterwards.
 *
 *     This reason will be used automatically.
 *
 * - CANCEL_ABORT / CANCEL_FAILED
 *     The user aborted the session.
 *
 *     If called from contactStarted1_ContactInvitesUser(),
 *     this sends an "MSNSLP 500 Internal Error" message, waits for
 *     the contact to ACK and terminates this application afterwards.
 *
 *     From other methods an P2P ACK message is sent with the "error flag" set (type 8).
 *
 *     Within the userStarted3_UserPrepares() method, CANCEL_ABORT is the only
 *     valid response to reject a data-preparation message. If KMess is compiled in debug mode,
 *     an assertion will be triggered otherwise.
 *
 * - CANCEL_NOT_INSTALLED
 *     The requested service/application is not installed.
 *
 *     This sends an "MSNSLP 500 Internal Error" message, waits for
 *     the contact to ACK and terminates this application afterwards.
 *
 *     This reason should only be used in contactStarted1_ContactInvitesUser().
 *     The default implementation in Application::contactStarted1_ContactInvitesUser()
 *     uses this reason automatically to cancel the invitation. Hence the base class
 *     is used to reject unknown services automatically.
 *
 * - CANCEL_TIMEOUT
 *     There was a timeout waiting for the contact to accept or send certain P2P data.
 *     Usually this reason is used internally by certain timeout functions.
 *     This sends an P2P message with the "waiting flag" set (type 4).
 *
 * - CANCEL_INVALID_SLP_CONTENT_TYPE
 *     Sends an "MSNSLP 500 Internal Error" message back, but without an Content-Type set.
 *
 *     This reason should only be used from contactStarted1_ContactInvitesUser()
 *     or userStarted2_ContactAccepts() to indicate the given Content-Type is not supported.
 *     Other invocations result in undefined responses from the other client (it will likely be ignored).
 *
 */
void P2PApplication::sendCancelMessage(const ApplicationCancelReason cancelReason)
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication - sendCancelMessage (reason=" << cancelReason << ")." << endl;
#endif

  // I don't care what the reason is if you had
  // to send the data-preparation ACK
  if(userShouldAcknowledge_)
  {
#ifdef KMESSTEST
    ASSERT( cancelReason == CANCEL_ABORT);
#endif

    // Cancelled the data preparation message
    sendP2PAck(P2PMessage::MSN_FLAG_ERROR);
    userShouldAcknowledge_ = false;
    shouldSendAck_         = false;
    // TODO: send bye here to? wait for bye? just terminate?
    kdWarning() << "P2PApplication::sendCancelMessage: data-preparation stage failed (contact=" << getContactHandle() << ")." << endl;
    endApplication();
    return;
  }


  switch(cancelReason)
  {
    // Declined the invitation
    case CANCEL_INVITATION:
#ifdef KMESSTEST
      ASSERT( ! invitationContentType_.isEmpty() );
      ASSERT( invitationSessionID_ != 0 );
#endif

      // Use the cancel-reason to send the correct message:
      sendSlpError("603 Decline", invitationSessionID_, invitationContentType_);
      return;


    // Timeout waiting for some P2P data
    // ...or timeout waiting for contact to accept.
    case CANCEL_TIMEOUT:
      // Timeout handling occurs via slotCleanup(), not here.
      kdWarning() << "P2PApplication: sendCancelMessage(CANCEL_TIMEOUT) is not supported "
                  << "(state=" << waitingState_ << " contact=" << getContactHandle() << ")" << endl;
      sendP2PMessage(0, P2PMessage::MSN_FLAG_WAITING, 0);
      shouldSendAck_ = false;
      endApplication();
      return;


    // The application type is not installed
    case CANCEL_NOT_INSTALLED:
      if( gotSlpMessage_)
      {
        // Usually this particular error is invoked from Application::contactStarted1_ContactInvitesUser().
        sendSlpError("500 Internal Error", invitationSessionID_, "null");
      }
      else
      {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
        kdWarning() << "P2PApplication: unexpected cancelReason used: CANCEL_NOT_INSTALLED." << endl;
#endif
      }
      return;


    // User or application wants to abort
    case CANCEL_ABORT:
    case CANCEL_FAILED:  // for compatibility with the MimeApplication class.
      // If we're sending data, abort that
      if( dataSource_ != 0 && dataType_ != P2P_TYPE_NEGOTIATION )
      {
        dataSource_ = 0;
        dataType_   = P2P_TYPE_NEGOTIATION;
      }

      // Determine which error message to send back.
      if(gotSlpMessage_)
      {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
        kdDebug() << "P2PApplication: Cancelling after receiving an SLP message, sending 500 Internal Error back." << endl;
#endif
        // Cancelled while in SLP negotiate mode.
        // Include the original content-type to tell the other
        // client that part of the message was not in error.
        sendSlpError("500 Internal Error", invitationSessionID_, invitationContentType_);
      }
      else if( isUserStartedApp() )
      {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
        kdDebug() << "P2PApplication: Cancelling an application we've started ourselves, sending BYE early." << endl;
#endif
        // Send bye early, contact should detect this.
        sendSlpBye();
      }
      else
      {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
        kdDebug() << "P2PApplication: Cancelling an application started by the other contact, sending BYE early." << endl;
#endif
        // Send bye early, contact should detect this.
        sendSlpBye();

        // TODO: after the message is ACKed, the contact sends a 0x40 "I've sent the bye early" message.
        // KMess does not do this here.
      }

      return;


    // Invalid content type received
    case CANCEL_INVALID_SLP_CONTENT_TYPE:
#ifdef KMESSTEST
      ASSERT( gotSlpMessage_ );
#endif
      // Send an 500 error message without content-type set
      sendSlpError("500 Internal Error");
      return;


    // Last option, unknown cancel message
    default:
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdWarning() << "P2PApplication: unknown cancelReason used." << endl;
#endif

      sendP2PAck(P2PMessage::MSN_FLAG_ERROR);
  }
}



/**
 * Send data packets from an data source. A timer/thread starts that transmits
 * all data packets automatically. It also uses the dataSource object
 * to determine the data size.
 *
 * @param  dataSource  Source of the data to send
 * @param  dataType    Type of the data, a member of the P2PDataType enum.
 */
void P2PApplication::sendData(QIODevice *dataSource, P2PDataType dataType)
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Begin of data transfer" << endl;
#endif
#ifdef KMESSTEST
  ASSERT( sessionID_ != 0 );
  ASSERT( dataSource != 0 );
#endif

  // Make sure we don't activate a slot without a data source.
  if(dataSource == 0)
  {
    kdWarning() << "P2PApplication: Couldn't send data, data source is null." << endl;


    showMessage( i18n("The transfer failed.  Couldn't open data source.") );

    // TODO: Official client sends BYE when data preparation failed.
    sendCancelMessage(CANCEL_FAILED); // MSNSLP messages terminate the application eventually.
    return;
  }

  // Set the global data fields
  dataSource_ = dataSource;
  dataType_   = dataType;

  // Set the total size field, this causes sendP2PMessage()
  // to handle fragmented messages correctly.
  messageTotalSize_ = dataSource->size();

  // Run the slot after 10ms
  QTimer::singleShot( 10, this, SLOT(slotSendData()) );
}



/**
 * Send the data preparation message.
 *
 * The data message consists of a P2P message with 0x00000000 as message body.
 */
void P2PApplication::sendDataPreparation()
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Sending data preparation message." << endl;
#endif
#ifdef KMESSTEST
  ASSERT( sessionID_    != 0 );
  ASSERT( waitingState_ == P2P_WAIT_DEFAULT || waitingState_ == P2P_WAIT_FOR_FILE_DATA );
#endif

  // TODO: Check the various footer codes send for certain transfers.
  QByteArray p2pMessage(4);
  p2pMessage.fill(0x00);
  sendP2PMessage(p2pMessage, 0, 1, P2P_MSG_DATA_PREPARATION);

  // Wait until our preparation message is ACK-ed
  waitingState_ = P2P_WAIT_FOR_PREPARE_ACK;
  waitingTimer_->start(30000, true); // 30 sec

  // TODO: after the data preparation is sent, WLM8 starts a direct connection setup:
  // - it does not send an ACK yet
  // - it sends the transreq message,
  // - when the invitation finished (or the DC is up..??) it sends the data preparation ACK.
  // - after that it sends the ACK of the last "SLP 200 OK" message.
  // - at this point, the data preparation ACK triggers KMess to send the picture data.

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Waiting for data preparation ACK..." << endl;
#endif
}



/**
 * A special method to handle certain acknowledgements.
 *
 * You need to call this method from userStarted3_UserPrepares()
 * to accept the data-preparation message.
 */
void P2PApplication::sendDataPreparationAck()
{
  if(userShouldAcknowledge_)
  {
    // This value is checked again after the
    // userStarted3_UserPrepares() method returns.
    userShouldAcknowledge_ = false;
  }
  else
  {
#ifdef KMESSTEST
    kdWarning() << "P2PApplication: Call to sendDataPreparationAck() not expected!" << endl;
#endif
  }
}



/**
 * Send the direct connection handshake
 */
void P2PApplication::sendDirectConnectionHandshake()
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Sending direct connection handshake (nonce=" << nonce_ << ")." << endl;
#endif
#ifdef KMESSTEST
  ASSERT( applicationList_->hasDirectConnection() );
#endif
#ifdef KMESSTEST
  ASSERT( waitingState_ == P2P_WAIT_FOR_CONNECTION2   // we just established connection, send the first handshake
       || waitingState_ == P2P_WAIT_FOR_HANDSHAKE     // contact established connection, waiting to receive first handshake
       || waitingState_ == P2P_WAIT_FOR_HANDSHAKE_OK  // we sent the first handshake, waiting for contact to return second.
       );
#endif

  // Message flag is 0x100
  const int handshakeFlag = P2PMessage::MSN_FLAG_DC_HANDSHAKE;

  // Prepare the header
  QByteArray header(48);
  header.fill(0x00);
  messageID_++;

  // Insert the bytes
  P2PMessage::insertBytes(header, 0,               0);   // 1: SessionID
  P2PMessage::insertBytes(header, messageID_,      4);   // 2: MessageID
  // offset and total size are 0
  P2PMessage::insertBytes(header, handshakeFlag,  28);   // 6: Message type (flags)

  // Insert the nonce at fields 7-9
  P2PMessage::insertNonce(header, nonce_);


#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  // Display binary data as hexadecimal string
  const int nonceOffset = 32;
  const char hexMap[] = "0123456789ABCDEF";
  QString hex;
  for(int i = 0; i < 16; i++)
  {
    if( i == 4 || i == 6 || i == 8 || i == 10 )
    {
      hex += "-";
    }
    int upper = (header[nonceOffset + i] & 0xf0) >> 4;
    int lower = (header[nonceOffset + i] & 0x0f);
    hex += hexMap[upper];
    hex += hexMap[lower];
  }
  kdDebug() << "P2PApplication: copied nonce '" << nonce_ << "' to p2p message header as '" << hex << "'." << endl;
#endif

  // Forcefully Send the message over the direct connection
  applicationList_->getDirectConnection()->sendMessage( header );

  if( waitingState_ != P2P_WAIT_FOR_HANDSHAKE_OK )
  {
    // Start the timeout timer again.
    waitingState_ = P2P_WAIT_FOR_HANDSHAKE_OK;
    waitingTimer_->start(30000, true); // 30 sec

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
    kdDebug() << "P2PApplication: Waiting for direct connection handshake response..." << endl;
#endif
  }
}



/**
 * Send a P2P ACK message.
 * MSN sends an P2P ACK for each received P2P message, which
 * turns the P2P communication into a stop-and-wait protocol.
 *
 * @param  ackType  The ACK type (flag) to use, defaults to MSN_FLAG_ACK.
 */
void P2PApplication::sendP2PAck(int ackType, UnAckedMessage *originalMessageData)
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  if(ackType == P2PMessage::MSN_FLAG_ACK)
  {
    if(messageID_ == 0)
    {
      kdDebug() << "P2PApplication: Sending P2P ACK (INVITE-ACK)" << endl;
    }
    else
    {
      kdDebug() << "P2PApplication: Sending P2P ACK" << endl;
    }
  }
  else if(ackType == P2PMessage::MSN_FLAG_ERROR)
  {
    kdDebug() << "P2PApplication: Sending P2P ACK (Error-response)" << endl;
  }
  else if(ackType == P2PMessage::MSN_FLAG_WAITING)
  {
    kdDebug() << "P2PApplication: Sending P2P ACK (Timeout-response)" << endl;
  }
  else if(ackType == P2PMessage::MSN_FLAG_CLOSING_ACK)
  {
    kdDebug() << "P2PApplication: Sending P2P ACK (Closing-ACK)" << endl;
  }
  else
  {
    kdDebug() << "P2PApplication: Sending P2P ACK (Unknown-type)" << endl;
  }
#endif
#ifdef KMESSTEST
  ASSERT( shouldSendAck_ );
#endif

  // Default to previous message if no argument is given
  if( originalMessageData == 0 )
  {
    originalMessageData = &lastIncomingMessage_;
  }


  // Fill the header:
  //
  // 0    4    8        16        24   28   32   36   40        48
  // |....|....|....|....|....|....|....|....|....|....|....|....|
  // |sid |mid |offset   |totalsize|size|flag|asid|auid|a-datasz |
  //

  QByteArray header(48);
  header.fill(0x00);

  // Determine the message ID
  if(messageID_ == 0)
  {
    // We don't have a message ID set yet, we're acknowledging
    // the first INVITE message. Generate an ID

    messageID_ = generateID();
    P2PMessage::insertBytes(header, messageID_, 4);

    // Somehow, the next message we send (200 OK) should contain a messageID - 3:
    // We decrement the messageID with 4 because it will be incremented in the next call.
    messageID_ -= 4;
  }
  else
  {
    messageID_++;
    P2PMessage::insertBytes(header, messageID_, 4);
  }


  // Insert the bytes
  P2PMessage::insertBytes(header, originalMessageData->sessionID,   0); // SessionID
  // Message ID was set earlier
  P2PMessage::insertBytes(header, originalMessageData->totalSize,  16); // Confirm total size set in previous message
  P2PMessage::insertBytes(header, ackType,                         28); // Set the ACK flag.
  P2PMessage::insertBytes(header, originalMessageData->messageID,  32); // Set the ACK message ID
  P2PMessage::insertBytes(header, originalMessageData->uniqueID,   36); // Set the ACK Unique ID field.
  P2PMessage::insertBytes(header, originalMessageData->dataSize,   40); // Set the ACK data size with the previous message size.


  // Footer
  char footer[4];
  footer[0] = 0x00;
  footer[1] = 0x00;
  footer[2] = 0x00;
  footer[3] = 0x00;


  // Deliver the message over the correct link (switchboard or direct connection)
  applicationList_->sendMessage(this, header);
  shouldSendAck_ = false;

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: ACK Transmitted (flags=0x" << QString::number(ackType, 16) << ")." << endl;
#endif
}



/**
 * Constructs and sends an P2P Message.
 *
 * This method includes some of the global fields in the message body.
 *
 * @param messageData    The message to send. This could be a binary file data block.
 * @param flagField      The flag-field content.
 * @param footerCode     The message footer.
 * @param messageType    The message type the message footprint is stored with in the unacked message queue.
 *                       This value is used later to respond to ACKs of certain special messages.
 */
void P2PApplication::sendP2PMessage(const QByteArray &messageData, int flagField, uint footerCode, P2PMessageType messageType)
{
#ifdef KMESSTEST
  ASSERT( messageData.size() <= 1202 );
#endif

  // Initialize the header
  QByteArray header(48);
  header.fill(0x00);


  // Determine the message ID:
  if(messageID_ == 0)
  {
    // This is the first P2P message we send, initialize message ID
    messageID_ = generateID();
  }
  else if(messageOffset_ == 0)
  {
    // Only increments the ID if this is not a "fragmented message".
    messageID_++;
  }


  // Verify we're not configured to send a splitted message.
  if( messageTotalSize_ > 0
  &&  flagField != 0
  &&  flagField != P2PMessage::MSN_FLAG_DATA
  &&  flagField != P2PMessage::MSN_FLAG_FILETRANSFER )
  {
    kdWarning() << "P2PApplication: Attempt to send splitted ACK message "
                << "(flags="   << flagField
                << " state="   << waitingState_
                << " contact=" << getContactHandle() << ")!" << endl;
  }


  // Determine the message size:
  uint messageSize = messageData.size();
  uint totalSize;
  uint offsetField;

  if(messageTotalSize_ == 0)
  {
    // Normal message:
    totalSize   = messageData.size();
    offsetField = 0;
  }
  else
  {
    // Splitted message:
    totalSize   = messageTotalSize_;
    offsetField = messageOffset_;
  }


  // Generate Unique ID for this message
  unsigned long uniqueID = generateID();


  // Update the header
  P2PMessage::insertBytes(header, sessionID_,     0);    // 1: Session ID
  P2PMessage::insertBytes(header, messageID_,     4);    // 2: BaseIdentifier (starts random [4, 4294967295])
  P2PMessage::insertBytes(header, offsetField,    8);    // 3: Offset (if more then 1202 bytes to send)
  P2PMessage::insertBytes(header, totalSize,      16);   // 4: Total size of the message
  P2PMessage::insertBytes(header, messageSize,    24);   // 5: This message size, size of previous if no data
  P2PMessage::insertBytes(header, flagField,      28);   // 6: Message type (flags)
  P2PMessage::insertBytes(header, uniqueID,       32);   // 7: Field2 of previous for ACK-messages (base message ID)
//P2PMessage::insertBytes(header, 0,              36);   // 8: Field7 of previous for ACK-messages (unique ID)
//P2PMessage::insertBytes(header, 0,              40);   // 9: Field4 of previous for ACK-messages (size)


  // Update unacked message  queue to handle incoming ACKs
  if( flagField == 0 || messageType != 0 )
  {
    if( offsetField == 0 )
    {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication: Storing message fields in queue for ACK response." << endl;
#endif
      // First message, or single message, store new record for ACKs.
      QDateTime currentTime;
      UnAckedMessage *ackData = new UnAckedMessage;
      ackData->dataSize     = messageSize;
      ackData->messageID    = messageID_;
      ackData->messageType  = messageType;  // used later to handle certain ACKs.
      ackData->sentTime     = currentTime.toTime_t();
      ackData->sessionID    = sessionID_;
      ackData->totalSize    = totalSize;
      ackData->uniqueID     = uniqueID;
      outgoingMessages_.append(ackData);
    }
    else
    {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication: Updating message fields in queue for ACK response." << endl;
#endif
      // Follow-up part of a splitted/fragmented message.
      // Update the uniqueID aka ackSID field to handle new acks.
      // TODO: it appears MSN7 uses the same ackSID for some messages, not sure
      //       whether that is a feature (to capture acks in a short time) or a problem in their random function.
      QPtrListIterator<UnAckedMessage> it(outgoingMessages_);
      while( it.current() != 0 )
      {
        if( it.current()->messageID == messageID_ )
        {
          it.current()->uniqueID = uniqueID;
          break;
        }
        ++it;
      }
    }
  }

  // Update the offset fields for splitted messages
  if(messageTotalSize_ != 0)
  {
    // Update offset
    messageOffset_ += messageSize;

    // If we transmitted all parts, reset the fields for the next normal message
    if(messageOffset_ >= messageTotalSize_)
    {
      messageOffset_    = 0;
      messageTotalSize_ = 0;
    }
  }


  // Deliver the message over the correct link (switchboard or direct connection)
  applicationList_->sendMessage(this, header, messageData, footerCode);


#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Message transmitted "
            << "(SID="         << sessionID_
            << " MID="         << messageID_
            << " UID="         << uniqueID
            << " flags=0x"     << QString::number(flagField, 16)
            << " size="        << messageSize
            << " type="        << messageType
            << ")" << endl;
#endif
}



/**
 * Close the session by sending a SLP BYE message.
 * This should only be called if we sent the INVITE as well.
 *
 * Once the BYE message is sent, it waits for the last ACK to arrive,
 * ACKs that message and terminates the application instance afterwards.
 *
 * If the last ACK doesn't arrive in a certain timeout, the application is also terminated.
 */
void P2PApplication::sendSlpBye()
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Sending SLP BYE message." << endl;
#endif
#ifdef KMESSTEST
  // Only send a BYE if we started it
  ASSERT( isUserStartedApp() );
#endif

  // Content-Type:
  QString contentType;
  contentType = "application/x-msnmsgr-sessionclosebody";

  // Handle
  QString myHandle = CurrentAccount::instance()->getHandle();


  // Create the message
  QString slpMessage;
  slpMessage = "BYE MSNMSGR:" + getContactHandle() + " MSNSLP/1.0\r\n"
               "To: <msnmsgr:" + getContactHandle() + ">\r\n"
               "From: <msnmsgr:" + myHandle + ">\r\n"
               "Via: MSNSLP/1.0/TLP ;branch=" + branch_ + "\r\n" // from our INVITE
               "CSeq: 0\r\n"
               "Call-ID: " + callID_ + "\r\n"  // from our INVITE
               "Max-Forwards: 0\r\n"
               "Content-Type: " + contentType + "\r\n"
               "Content-Length: 3\r\n"
               "\r\n";

  // Send the message
  sendSlpMessage(slpMessage, P2P_MSG_SESSION_BYE);
  setClosing(true);

  // Don't run endApplication() yet, there is one ACK we wait for...
  waitingState_ = P2P_WAIT_FOR_SLP_BYE_ACK;
  waitingTimer_->start(30000, true); // 30 sec

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Waiting for BYE ACK..." << endl;
#endif
}



/**
 * Send an error response at SLP level.
 * This can be done to decline a file tranfer or tell the other client it sent bad SLP data.
 *
 * @param  statusLine          The status line, for example "200 OK", or "603 Decline"
 * @param  sessionID           ID of the current session (can be empty)
 * @param  messageContentType  Contact type of the message which is in error (can be empty)
 *
 * The session ID can be 0 to indicate there was an error in the Content-Type. Setting the
 * session ID and Content-Type to the values used in the INVITE message indicates
 * you like to cancel that invitation.
 *
 * Once this method returns the SessionID is set to 0 (which shouldn't be a problem
 * since you use this method to end a session). This is because SLP messages are always
 * sent without an SessionID.
 */
void P2PApplication::sendSlpError(const QString &statusLine,
                                  const uint sessionID, const QString &messageContentType, P2PMessageType messageType)
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Sending SLP error response: " << statusLine << endl;
#endif

#ifdef KMESSTEST
  // Test, as other invocations likely result in undefined client behavour
  // (
  ASSERT( ! shouldSendAck_ );
  ASSERT(   gotSlpMessage_ );
  ASSERT( waitingState_ == P2P_WAIT_DEFAULT );
  ASSERT( ! callID_.isEmpty() );
  ASSERT( ! branch_.isEmpty() );
#endif

  QString myHandle = CurrentAccount::instance()->getHandle();

  // Determine Content-Type and Content-Length fields
  QString content;
  QString contentType;
  uint    contentLength;

  if( callID_.isEmpty() || branch_.isEmpty() )
  {
    kdWarning() << "P2PApplication: attempting to send SLP error back, but Call-ID or Branch is unknown! "
                << "Sending 0x08 error instead (contact=" << getContactHandle() << ")" << endl;
    sendP2PAck(P2PMessage::MSN_FLAG_ERROR);
    return;
  }

  if(sessionID == 0)
  {
    // Used to indicate a general error
    content       = QString::null;
    contentType   = "null";
    contentLength = 0;
  }
  else
  {
#ifdef KMESSTEST
    ASSERT(messageContentType != 0 && ! messageContentType.isEmpty());
#endif
    // Used to indicate an error in the Invite, or decline a session
    content       = "SessionID: " + QString::number(sessionID) + "\r\n\r\n\0";
    contentLength = content.length();
    contentType   = messageContentType;
  }

  // Create the message
  QString slpMessage;
  slpMessage = "MSNSLP/1.0 " + statusLine + "\r\n"
               "To: <msnmsgr:" + getContactHandle() + ">\r\n"
               "From: <msnmsgr:" + myHandle + ">\r\n"
               "Via: MSNSLP/1.0/TLP ;branch=" + branch_ + "\r\n"  // Indicates which INVITE was in error.
               "CSeq: 0\r\n"
               "Call-ID: " + callID_ + "\r\n"
               "Max-Forwards: 0\r\n"
               "Content-Type: " + contentType + "\r\n"
               "Content-Length: " + QString::number(contentLength) + "\r\n"
               + (contentLength > 0 ? "\r\n" : "")
               + content;

  // Send the message
  sendSlpMessage(slpMessage, messageType);
  setClosing(true);

  // Wait for the message to be ACK-ed.
  waitingState_ = P2P_WAIT_FOR_SLP_ERR_ACK;
  waitingTimer_->start(30000, true); // 30 sec

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Waiting for SLP ACK..." << endl;
#endif
}



/**
 * Start a new MSNSLP session
 *
 * @param  message      Invitation message body.
 * @param  contentType  Content-Type of the message body.
 */
void P2PApplication::sendSlpInvitation(const MimeMessage &message, const QString &contentType)
{
#ifdef KMESSTEST
  ASSERT(! contentType.isEmpty()    );
  ASSERT( waitingState_ == P2P_WAIT_DEFAULT );
#endif

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Sending INVITE message" << endl;
#endif

  // Mime message conversion
  QString mimeBody = message.getFields();

  // Content handles
  QString myHandle = CurrentAccount::instance()->getHandle();

  // Generate identifiers
  if( callID_.isNull() )
  {
    // The callID remains the same between multiple invites
    // It identifies the session.
    callID_ = generateGUID();
  }

  // BranchID identifies this INVITE request
  branch_ = generateGUID();

  // Create the message (space after CSeq is added to emulate msn exactly)
  QString slpMessage;
  slpMessage = "INVITE MSNMSGR:" + getContactHandle() + " MSNSLP/1.0\r\n"
               "To: <msnmsgr:" + getContactHandle() + ">\r\n"
               "From: <msnmsgr:" + myHandle + ">\r\n"
               "Via: MSNSLP/1.0/TLP ;branch=" + branch_ + "\r\n"
               "CSeq: 0 \r\n"
               "Call-ID: " + callID_ + "\r\n"
               "Max-Forwards: 0\r\n"
               "Content-Type: " + contentType + "\r\n"   // Starts MSNSLP Session.
               "Content-Length: " + QString::number(mimeBody.length() + 5) + "\r\n"
               "\r\n"
               + mimeBody +
               "\r\n";

  // Send the message
  messageOffset_    = 0;
  messageTotalSize_ = 0;

  sendSlpMessage(slpMessage, P2P_MSG_SESSION_INVITATION);

  // Keep the used content type in the state variables.
  // For contact-started invitations, this is used to store their type.
  // In this case, it can be used to store ours.
  invitationContentType_ = contentType;

  // Wait for the contact to accept
  waitingState_     = P2P_WAIT_CONTACT_ACCEPT;
  waitingTimer_->start(30000, true); // 30 sec (MSN6 also has a 30 sec limit for this)

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Waiting for contact to accept..." << endl;
#endif
}



/**
 * Send a given string using sendP2PMessage().
 * The SLP message is always sent with a sessionid of zero.
 */
void P2PApplication::sendSlpMessage(const QString &slpMessage, P2PMessageType messageType)
{
  // Make sure the message is sent with sessionid=0 (for BYE messages)
  unsigned long oldSessionID = sessionID_;
  sessionID_ = 0;

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication::sendSlpMessage: SLP message=" << slpMessage << endl;
#endif

  // Get the raw utf8 encoded bytearray
  // This converts the QCString to QByteArray directly.
  // Don't try to remove the extra padded '\0' character
  // (like MimeMessage::getMessage() does)
  // because MSN6/7 actually relies on this.
  QByteArray utf8Message = slpMessage.utf8();
  QByteArray messagePart;

  // Make sure the message gets sent in chunks of 1202 bytes
  messageOffset_      = 0;
  messageTotalSize_   = utf8Message.size();
  uint remainingBytes = messageTotalSize_;

  while(remainingBytes > 0)
  {
    // Let another QByteArray encapsulate a part of the utf8Message
    char *dataPointer = utf8Message.data() + messageOffset_;
    uint  dataSize    = QMIN(remainingBytes, 1202);
    messagePart.setRawData( dataPointer, dataSize );

    // Send the message with the generic sendP2PMessage() method
    // messageOffset_ is updated automatically
    sendP2PMessage(messagePart, 0, 0, messageType);
    remainingBytes -= dataSize;

#ifdef KMESSTEST
    ASSERT( remainingBytes == messageTotalSize_ - messageOffset_ );
#endif

    // Release data
    messagePart.resetRawData( dataPointer, dataSize );
  }

  // Reset for normal operations
  messageOffset_    = 0;
  messageTotalSize_ = 0;

  // Restore the session ID
  sessionID_ = oldSessionID;
}



/**
 * Send the 200 OK message to accept the invitation.
 * Most likely the content message should contain the SessionID sent in the invitation.
 * Once the confirmation message has been sent, this class will automatically
 * use the correct Session ID (because it was read when the invite was received).
 *
 * @param message  Body of the SLP message
 */
void P2PApplication::sendSlpOkMessage(const MimeMessage &message)
{
#ifdef KMESSTEST
  ASSERT(! branch_.isEmpty()                );
  ASSERT(! invitationContentType_.isEmpty() );
  ASSERT( waitingState_ == P2P_WAIT_DEFAULT );  // TODO: WLM8 invitations before data preparation
#endif

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Sending MSNSLP/1.0 200 OK message." << endl;
#endif

  // Get message content and Content-Type
  bool gotTransferInvitation = (invitationContentType_ == "application/x-msnmsgr-transreqbody");
  QString content     = message.getFields();
  QString contentType = (gotTransferInvitation ? "application/x-msnmsgr-transrespbody" : invitationContentType_);

  QString myHandle = CurrentAccount::instance()->getHandle();

  // Send the SLP OK message to tell the other client you'd like to start the transfer.
  QString slpMessage;
  slpMessage = "MSNSLP/1.0 200 OK\r\n"
               "To: <msnmsgr:" + getContactHandle() + ">\r\n"
               "From: <msnmsgr:" + myHandle + ">\r\n"
               "Via: MSNSLP/1.0/TLP ;branch=" + branch_ + "\r\n"
               "CSeq: " + QString::number(invitationCSeq_ + 1) + "\r\n"
               "Call-ID: " + callID_ + "\r\n"
               "Max-Forwards: 0\r\n"
               "Content-Type: " + contentType + "\r\n"
               "Content-Length: " + QString::number(content.length() + 3) + "\r\n"
               "\r\n"
               + content + "\r\n";

  sendSlpMessage(slpMessage, ( gotTransferInvitation ? P2P_MSG_TRANSFER_OK : P2P_MSG_SESSION_OK ) );


  // Make sure it's not overwritten when the other contact invites for a DC transfer.
  if( sessionID_ == 0 )
  {
    // Now that we've confirmed the session, use the sessionID
    // in the next messages:
    sessionID_ = invitationSessionID_;
  }



  if(gotTransferInvitation)
  {
    // Wait until our transfer OK message is ACK-ed
    waitingState_ = P2P_WAIT_FOR_TRANSFER_ACK;
    waitingTimer_->start(30000, true); // 30 sec

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
    kdDebug() << "P2PApplication: Waiting for contact to ACK the 200 transfer OK message..." << endl;
#endif
  }
  else
  {
    // Wait until our normal OK message is ACK-ed
    waitingState_ = P2P_WAIT_FOR_SLP_OK_ACK;
    waitingTimer_->start(30000, true); // 30 sec

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
    kdDebug() << "P2PApplication: Waiting for contact to ACK the 200 OK message..." << endl;
#endif
  }
}



// Send the invitation for a normal session
void P2PApplication::sendSlpSessionInvitation(uint sessionID, const QString &eufGuid, const int appId, const QString &context)
{
  // Create the message
  MimeMessage invitation;
  invitation.addField( "EUF-GUID",  eufGuid );
  invitation.addField( "SessionID", QString::number(sessionID) );
  invitation.addField( "AppID",     QString::number(appId)     );
  invitation.addField( "Context",   context                    );

  // Send the invitation
  sendSlpInvitation( invitation, "application/x-msnmsgr-sessionreqbody");

  // If the client accepts our invitation, we
  // use the session ID in the next ACK message.
  sessionID_        = sessionID;
}



// Send the invitation for a direct connection.
void P2PApplication::sendSlpTransferInvitation()
{
  // Check if there is already an connection
  if( applicationList_->hasDirectConnection() )
  {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
    kdDebug() << "P2PApplication::sendSlpSessionInvitation: Connection is already setup, re-using the current connection." << endl;
#endif
    initiateTransfer();
    return;
  }

  // Prepare to transfer when a connection is made by this application, or another application.
  connect( applicationList_, SIGNAL(      connectionAuthorized() ),   // when authorized, send data
           this,             SLOT  (  slotConnectionAuthorized() ));
  connect( applicationList_, SIGNAL(          connectionFailed() ),   // when failed, revert to switchboard
           this,             SLOT  (      slotConnectionFailed() ));

  // Check if there are pending connections
  if( applicationList_->hasPendingConnections() || applicationList_->hasPendingConnectionInvitation() )
  {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
    kdDebug() << "P2PApplication::sendSlpSessionInvitation: Another application is already inviting for a direct connection, "
              << "waiting for a connection to succeed or fail." << endl;
#endif

    showTransferMessage( i18n("File transfer dialog message", "Waiting for connection") );
    return;
  }

  // Only connect last slot if we're really going to make the connection ourselves.
  connect( applicationList_, SIGNAL(     connectionEstablished() ),   // when connected, send nonce
           this,             SLOT  ( slotConnectionEstablished() ));

  // Invite the contact to create a connection.
  // This message tells the other client it can't connect to this end-host.
  // Either the data needs to be sent over the switchboard,
  // or the other client should initiate a direct connection server.
  MimeMessage invitation;
  invitation.addField("Bridges",      "TCPv1");
  invitation.addField("NetID",        QString::number(generateID()));
  invitation.addField("Conn-Type",    "IP-Restrict-NAT");
  invitation.addField("UPnPNat",      "false");
  invitation.addField("ICF",          "false");

  // Send the invitation
  sendSlpInvitation( invitation, "application/x-msnmsgr-transreqbody" );
  waitingState_ = P2P_WAIT_TRANSFER_ACCEPT;

  // Mark invitation as pending.
  // This will fire an event when it's manually reset to false.
  applicationList_->setPendingConnectionInvitation(true);


  // When we include a Hashed-Nonce field, the contact also
  // adds a Hashed-Nonce to it's 'transrespbody' response message.
  // This may resemble http-digest authentication, but is still not researched.

  // Supported Conn-Type values:
  // - Firewall
  // - Direct-Connect     (Same IP, same port)
  // - Port-Restrict-NAT  (Same IP, different port)
  // - IP-Restrict-NAT    (Different IP, same port)
  // - Symmetric-NAT      (Different IP, different port)
  // - Unknown-Connect    (Different IP, different port)
  // The connection type can be discovered with STUN.

  // NetID is zero for Firewall/Direct-Connect,
  // otherwise it's a random number.

  // Inform the user about the file transfer progress.
  showTransferMessage( i18n("File transfer dialog message", "Negotiating options to connect") );
}



/**
 * Cleanup function, is called if we never received the data we were waiting for.
 */
void P2PApplication::slotCleanup()
{
#ifdef KMESSTEST
  ASSERT( waitingState_ != P2P_WAIT_DEFAULT );
#endif

  // Not waiting for ourselves to connect.
  if( waitingState_ != P2P_WAIT_FOR_CONNECTION2 )
  {
    if( aborting_ )
    {
      // Don't send messages when aborting. This could potentially crash KMess, or cause new windows to appear.
      // - suppose this is the last aborting application which requests removal.
      // - ApplicationList informs the MsnSwitchboardConnection it can remove itself.
      // - ChatMaster processes the putMsg() call, and attempts to deliver the message.
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication::slotCleanup: Not sending error message because application is aborting." << endl;
#endif
    }
    else
    {
      // Tell the other client we were waiting for a message to receive!
      sendP2PMessage(0, P2PMessage::MSN_FLAG_WAITING, 0);
    }
  }

  // Important note:
  // Each case calling endApplication() contains a "return;"
  // All debug cases use "break;" so the final endApplication() is called.

  // Put a message at the console
  switch(waitingState_)
  {
    case P2P_WAIT_CONTACT_ACCEPT:
    {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication: Contact didn't accept the invitation, terminating manually." << endl;
#endif

      endApplication( i18n("The invitation was cancelled.  A timeout occured waiting for the contact to accept.") );
      return;
    }

    case P2P_WAIT_FOR_CONNECTION:
    {
      endApplication( i18n("The invitation was cancelled.  A timeout occured waiting for a connection to succeed or fail.") );
      break;
    }

    case P2P_WAIT_FOR_CONNECTION2:
    {
      // Still keep waiting, DirectConnectionBase should trigger a timeout.
      kdWarning() << "P2PApplication: Time-out waiting to connect to the remote system (contact=" << getContactHandle() << ")." << endl;
      return;
    }

    case P2P_WAIT_FOR_SLP_OK_ACK:
    {
      kdWarning() << "P2PApplication: Timeout waiting for SLP OK ACK (contact=" << getContactHandle() << ")." << endl;
      break;
    }

    case P2P_WAIT_FOR_PREPARE:
    {
      kdWarning() << "P2PApplication: Timeout waiting for data-preparation message (contact=" << getContactHandle() << ")." << endl;
      break;
    }

    case P2P_WAIT_FOR_PREPARE_ACK:
    {
      kdWarning() << "P2PApplication: Timeout waiting for data-preparation ACK (contact=" << getContactHandle() << ")." << endl;
      break;
    }

    case P2P_WAIT_FOR_TRANSFER_ACK:
    {
      kdWarning() << "P2PApplication: Timeout waiting for SLP transfer OK ACK (contact=" << getContactHandle() << ")." << endl;
      break;
    }

    case P2P_WAIT_FOR_FILE_DATA:
    {
      kdWarning() << "P2PApplication: Timeout waiting for file data transfer (contact=" << getContactHandle() << ")." << endl;
      break;
    }

    case P2P_WAIT_FOR_DATA_ACK:
    {
      kdWarning() << "P2PApplication: Timeout waiting for file data received ack (contact=" << getContactHandle() << ")." << endl;
      break;
    }

    case P2P_WAIT_FOR_HANDSHAKE:
    {
      // NOTE: we could start an initiateTransfer() here, but this is clearly a client bug.
      kdWarning() << "P2PApplication: Timeout waiting for DC handshake (contact=" << getContactHandle() << ")." << endl;
      break;
    }
    
    case P2P_WAIT_FOR_HANDSHAKE_OK:
    {
      // NOTE: we could start an initiateTransfer() here, but this is clearly a client bug.
      kdWarning() << "P2PApplication: Timeout waiting for DC handshake response (contact=" << getContactHandle() << ")." << endl;
      break;
    }

    case P2P_WAIT_FOR_SLP_BYE:
    {
      kdWarning() << "P2PApplication: Timeout waiting for BYE message (contact=" << getContactHandle() << ")." << endl;
      endApplication(); // No message required, already got what we wanted
      return;
    }

    case P2P_WAIT_FOR_SLP_BYE_ACK:
    {
      kdWarning() << "P2PApplication: Timeout waiting for BYE-ACK (contact=" << getContactHandle() << ")." << endl;
      endApplication(); // No message required, already got what we wanted
      return;
    }

    case P2P_WAIT_FOR_SLP_ERR_ACK:
    {
      kdWarning() << "P2PApplication: Timeout waiting for SLP Error-ACK (contact=" << getContactHandle() << ")." << endl;
      endApplication(); // No message required, already got what we wanted
      return;
    }

    default:
    {
      kdWarning() << "P2PApplication: Timeout waiting (state=" << waitingState_ << " contact=" << getContactHandle() << ")" << endl;
    }
  }

  endApplication( i18n("The invitation was cancelled.  A timeout occured waiting for data.") );
}



/**
 * The direct connection is authorized by one of the P2P applications.
 * Each pending application can start the transfer now.
 */
void P2PApplication::slotConnectionAuthorized()
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Direct connection authorized, starting transfer." << endl;
#endif

  // No longer wait for packets, got connection.
  waitingTimer_->stop();

  initiateTransfer();
}



/**
 * The direct connection is established.
 * Initialize the connection, send the handshake to authorize it.
 */
void P2PApplication::slotConnectionEstablished()
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Direct connection established." << endl;
#endif

  // Disconnect irrelevant signals.
  // Connection can still fail if the nonce is invalid.
  disconnect( this, SLOT( slotConnectionEstablished() ));

  // Send the handshake if this is a new connection
  bool isClient = (! applicationList_->getDirectConnection()->isServer());
  if( isClient )
  {
    sendDirectConnectionHandshake();
  }
  else
  {
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
    kdDebug() << "P2PApplication: Waiting for contact to send connection handshake..." << endl;
#endif
    waitingState_ = P2P_WAIT_FOR_HANDSHAKE;
    waitingTimer_->start(30000, true); // 30 sec
  }
}



/**
 * The direct connection was closed
 */
void P2PApplication::slotConnectionClosed()
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Direct connection closed, STUB." << endl;
#endif

  // Disconnect irrelevant signals.
  // connectionEstablished was disconnected before.
  disconnect( this, SLOT( slotConnectionAuthorized()  ));
  disconnect( this, SLOT( slotConnectionFailed()      ));

  // TODO: now what..?
  // Note that ApplicationList::setPendingConnectionInvitation() also triggers a connectionFailed signal.
}



/**
 * A direct connection could not be made.
 */
void P2PApplication::slotConnectionFailed()
{
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: The direct connection could not be made, transfer continues over the switchboard." << endl;
#endif
#ifdef KMESSTEST
  ASSERT( ! applicationList_->hasDirectConnection()   );
  ASSERT( ! applicationList_->hasPendingConnections() );
#endif

  // Disconnect all signals as they are not relevant anymore
  disconnect( this, SLOT( slotConnectionAuthorized()  ));
  disconnect( this, SLOT( slotConnectionClosed()      ));
  disconnect( this, SLOT( slotConnectionEstablished() ));
  disconnect( this, SLOT( slotConnectionFailed()      ));

  // Inform the user about the transfer progress
  showTransferMessage( i18n("Reverting to indirect file transfer (this could be slow).") );

  // start the transfer over the switchboard
  waitingTimer_->stop();
  initiateTransfer();
}



/**
 * A slot which sends the file data.
 * This slot is called from a timer event.
 */
void P2PApplication::slotSendData()
{
  // KMess would crash otherwise...
  if( dataSource_ == 0 )
  {
    kdDebug() << "P2PApplication: WARNING - Couldn't send more data, data source is null." << endl;

    // Don't send an error message, already happened before.
    return;
  }

  // Read the data from the file.
  char buffer[1202]; // 1202 is the max P2P data size.
  int bytesRead = dataSource_->readBlock( buffer, 1202 );

  QByteArray data;
  data.duplicate( buffer, bytesRead );

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
  kdDebug() << "P2PApplication: Sending data:"
            << " offset=" << messageOffset_
            << " buffer=" << bytesRead
            << endl;
#endif

  // Dispatch the progress to the derived class
  // Note that the messageOffset_ field is reset by sendP2PMessage().
  showTransferProgress( messageOffset_ + bytesRead );

  // Determine the flags:
  // Funny, MSN6 didn't seam to care which flag was set
  // while transferring the file. KMess however, does.
  if( dataType_ == P2P_TYPE_PICTURE )
  {
    sendP2PMessage(data, P2PMessage::MSN_FLAG_DATA, 1, P2P_MSG_DATA);
  }
  else if( dataType_ == P2P_TYPE_FILE )
  {
    sendP2PMessage(data, P2PMessage::MSN_FLAG_FILETRANSFER, 2, P2P_MSG_DATA);
  }
  else
  {
    if(messageOffset_ == 0) kdDebug() << "P2PApplication: Unknown data transfer mode." << endl;
    sendP2PMessage(data, 0, 1, P2P_MSG_DATA); // The best solution in this situation
  }

  // Finished?
  if(messageTotalSize_ > 0)
  {
    bool hasDirectConnection = applicationList_->hasDirectConnection();

    // Not finished
    if( ! hasDirectConnection && isTransferPaused() )
    {
      // The pause functionality is used when the switchboard has way too much messages unacked.
      // If we continue to queue messages, it will start to consume all memory with the file contents.
#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
      kdDebug() << "P2PApplication: Switchboard requested to pause the transfer." << endl;
#endif
    }
    else
    {
      // Run this slot again, transfer rate depends whether we're using a switchboard or not.
      // The MsnSwitchboardConnection/ChatMaster class queue messages when there are too many unacked ones.
      // This avoids flooding the client memory. The switchboard server won't accept more messages anyway.
      int timeout = ( hasDirectConnection ? 10 : 100 );   // 100 ms: 1200 bytes * (1000/50) = 14kb/s max.
      QTimer::singleShot( timeout, this, SLOT(slotSendData()) );
    }
  }
  else
  {
    // Transfer finished!
#ifdef KMESSTEST
    ASSERT( dataSource_->atEnd() );
#endif

    // Reset data source
    dataSource_   = 0;
    dataType_     = P2P_TYPE_NEGOTIATION;

#ifdef KMESSDEBUG_P2PAPPLICATION_GENERAL
    kdDebug() << "P2PApplication: Finished sending data, waiting for ACK..." << endl;
#endif

    waitingState_ = P2P_WAIT_FOR_DATA_ACK;
    waitingTimer_->start(30000, true); // 30 sec
  }
}



/**
 * You have cancelled the session.
 */
void P2PApplication::userAborted()
{
#ifdef KMESSDEBUG_APPLICATION
  kdDebug() << "Application::userAborted" << endl;
#endif

  aborting_ = true;
  setClosing(true);
  showMessage( getUserAbortMessage() );
  sendCancelMessage( CANCEL_ABORT );

  // Don't end the application, instead wait for contact's client to respond too
}



/**
 * You have rejected the invitation.
 */
void P2PApplication::userRejected()
{
#ifdef KMESSDEBUG_APPLICATION
  kdDebug() << "Application::userRejected" << endl;
#endif

  aborting_ = true;
  setClosing(true);
  showMessage( getUserRejectMessage() );
  sendCancelMessage( CANCEL_INVITATION );

  // Don't end the application, instead wait for contact's client to respond too
}


#include "p2papplication.moc"
