/***************************************************************************
                          httpsoapconnection.h -  description
                             -------------------
    begin                : Sun Sep 25 2005
    copyright            : (C) 2005 by Diederik van der Boor
    email                : vdboor --at-- codingdomain.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef HTTPSOAPCONNECTION_H
#define HTTPSOAPCONNECTION_H

#include "../tcpconnectionbase.h"

#include <kurl.h>

#include <qdom.h>

class SoapMessage;
class MimeMessage;


/**
 * @brief HTTP connection class to transport SOAP messages.
 * @author Diederik van der Boor
 * @ingroup NetworkExtra
 *
 * The class implements SOAP at a basic level; much XmlHttpRequest in web browsers.
 * This allows you to submit sniffed-data through sendRequest() without much hassle.
 * The result nodes are returned as QDomElement, allowing traversing through the returned data structures.
 *
 * The class can be used externally by connecting to it's signals,
 * or extended by a derived class instead.
 */
class HttpSoapConnection : public TcpConnectionBase
{
  Q_OBJECT

  public:  // public methods
    // The constructor
                         HttpSoapConnection( const QString &endpoint, QObject *parent = 0, const char *name = "HttpSoapConnection" );
    // The destructor
    virtual             ~HttpSoapConnection();

    // Close the connection
    virtual void         closeConnection();

    // Return the endpoint
    QString              getEndpoint() const;
    // Whether the connection is idle, not processing a SOAP request/response
    bool                 isIdle();
    // Return the received HTTP status code
    int                  getHttpStatusCode() const;
    // Return the last request ID
    const QString &      getRequestId() const;
    // Return the last SOAP action
    const QString &      getSoapAction() const;

    // Send a SOAP request
    void                 sendRawPostRequest( const QString &content, const QString &soapAction = QString::null, const QString &requestId = QString::null );
    // Send a SOAP request
    void                 sendRequest( const QString &soapAction, const SoapMessage &message, const QString &requestId = QString::null );
    // Send a SOAP request
    void                 sendRequest( const QString &soapAction, const QString &messageBody, const QString &messageHeader = QString::null, const QString &requestId = QString::null );

  protected:
    // Escape a string for inclusion in XML
    QString              escapeString( QString value ) const;
    // Parse the HTTP header
    virtual bool         parseHttpHeader( const QString &preamble, const MimeMessage &header );
    // The connection received the full response
    virtual bool         parseHttpBody( const QByteArray &responseBody );
    // The connection failed
    virtual void         slotConnectionFailed();
    // The connection's request failed
    virtual void         slotRequestFailed( const QString &faultCode, const QString &faultString, QDomElement &faultNode );
    // The connection's request completed
    virtual void         slotRequestFinished( QDomElement &resultRoot );

  private:
    // Connect to the endpoint
    bool                 connectToEndpoint();
    // The connection is established
    void                 slotConnectionEstablished();
    // The connection received data
    void                 slotDataReceived();

  private:  // private attributes
    // The endpoint for the request
    KURL                 endpoint_;
    // Whether the server requests to keep the connection alive
    bool                 keepAlive_;
    // Whether there is a parsing error.
    bool                 parseError_;
    // The message to be sent
    QCString             request_;
    // Whether there is an request active.
    bool                 requestActive_;
    // The remote ID of the request.
    QString              requestId_;
    // The response code
    uint                 responseCode_;
    // The length of the response
    uint                 responseLength_;
    // The last SOAP action
    QString              soapAction_;

  signals:
    // Called when another error occured
    void                 requestFailed( HttpSoapConnection *client, const QString &reason );
    // Called when a http request failed
    void                 requestFailed( HttpSoapConnection *client, const QString &faultCode,
                                        const QString &faultString, QDomElement faultNode );
    // Called when a query completed
    void                 requestFinished( HttpSoapConnection *client, QDomElement resultRoot );
};

#endif
